/****************************************************************************
**
** Copyright (C) 2006-2008 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
**
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qdocument.h"

/*
	Document model :
	
	Goals :
		* KISS : Keep It Simple Stupid
		* FAST : ...
		* LIGHTWEIGHT : reduce memory usage
		* FLEXIBLE : allow punctual bidi through QTextLayout
		
	Implementation :
		QDocument
		QDocumentPrivate
		
		QDocumentLine and QDocumentLineHandle => equivalent of QTextBlock
		QDocumentCursor and QDocumentCursorHandle => equivalent of QTextCursor
	
	Note :
		The KISS principle has been kinda mistreated in the private API due to
		the addition of some complex features which where not planned to be
		supported when defining the goals (e.g line wrapping, variable width
		fonts, ...). Such a compromission is affordable but should be avoided
		whenever possible in the future. And of course the public API should
		never suffer from such a thing.
*/

/*!
	\ingroup document
	@{
*/

/*!
	\class QDocument
	
	\brief A class storing a document
	
	QCE uses an architecture very similar to that of QTextEdit/QTextDocument
	which closely ressembles model/view. The document holds all the textual
	and formatting data. It offers some (mostly indirect) ways of modifying
	its content and is usable without any GUI.
	
	In QCE, a document is merely a list of QDocumentLine objects on which some
	extra formatting can be applied and which can be wrapped/hidden in various
	ways.
	
	The document model has been designed with three goals in mind :
	<ul>
	<li>performance
	<li>flexibility
	<li>low memory usage
	</ul>
	
	QDocument supports Bidi by using QTextLayout on lines that require it and
	prefers custom rendering in other cases to achieve the above goals.
	
	All the actual text editing is done through QDocumentCursor objects.
	
	\see QDocumentLine
	\see QDocumentCursor
*/

#include "qdocument_p.h"
#include "qdocumentmark.h"
#include "qdocumentcommand.h"
#include "qdocumentregiontree.h"

#include "qformat.h"
#include "qformatscheme.h"
#include "qlanguagedefinition.h"
#include "qmarksmanager.h"

#include <QPen>
#include <QTime>
#include <QRect>
#include <QLine>
#include <QDebug>
#include <QPainter>
#include <QPrinter>
#include <QFileInfo>
#include <QTextStream>
#include <QTextLayout>
#include <QApplication>
#include <QVarLengthArray>
#include <QMessageBox>

static int m_spaceSignOffset = 2;

static QPoint m_spaceSign[] = {
	QPoint(2, -1),
	QPoint(2, 0),
	QPoint(3, 0)
};

inline static bool isWord(QChar c)
{ return c.isLetterOrNumber() || (c == QLatin1Char('_')); }

int QDocument::screenLength(const QChar *d, int l, int tabStop)
{
	if ( tabStop == 1 )
		return l;
	
	int idx, column = idx = 0;
	
	while ( idx < l )
	{
		QChar c = d[idx];
		
		if ( c == QLatin1Char('\t') )
		{
			int taboffset = tabStop - (column % tabStop);
			column += taboffset;
		} else {
			++column;
		}
		
		++idx;
	}
	
	//qDebug("%s : %i", qPrintable(QString(d, l)), column);
	
	return column;
}

QString QDocument::screenable(const QChar *d, int l, int tabStop)
{
	if ( tabStop == 1 )
		return QString(d, l);
	
	QString fragment;
	int idx, column = idx = 0;
	
	while ( idx < l )
	{
		QChar c = d[idx];
		
		if ( c == QLatin1Char('\t') )
		{
			int taboffset = tabStop - (column % tabStop);
			
			fragment += QString(taboffset, QLatin1Char(' '));
			column += taboffset;
		} else {
			fragment += c;
			++column;
		}
		
		++idx;
	}
	
	return fragment;
}

struct InitStruct
{
	InitStruct()
	{
		
	}
};

static InitStruct init_inst;

/*!
	\brief ctor
*/
QDocument::QDocument(QObject *p)
 : QObject(p), m_impl(new QDocumentPrivate(this))
{
	if ( !QDocumentPrivate::m_font )
	{
		// must not happen if config dialog plugged in...
		setFont(QFont("Monospace", 10));
	}
	
	setText(QString());
	setLineEnding(QDocument::Conservative);
	
	connect(&(m_impl->m_commands)	, SIGNAL( cleanChanged(bool) ),
			this					, SIGNAL( cleanChanged(bool) ) );
	
	connect(&(m_impl->m_commands)	, SIGNAL( canUndoChanged(bool) ),
			this					, SIGNAL( undoAvailable(bool) ) );
	
	connect(&(m_impl->m_commands)	, SIGNAL( canRedoChanged(bool) ),
			this					, SIGNAL( redoAvailable(bool) ) );
	
}

/*!
	\brief dtor
*/
QDocument::~QDocument()
{
	delete m_impl;
}

/*!
	\brief Clear the content of the document
*/
void QDocument::clear()
{
	setText(QString());
}

/*!
	\return whether there commands to undo on the command stack
*/
bool QDocument::canUndo() const
{
	return m_impl ? m_impl->m_commands.canUndo() : false;
}

/*!
	\return whether there are commands to redo on the command stack
*/
bool QDocument::canRedo() const
{
	return m_impl ? m_impl->m_commands.canRedo() : false;
}

/*!
	\brief Undo the last editing operation
*/
void QDocument::undo()
{
	if ( m_impl )
		m_impl->m_commands.undo();
	
}

/*!
	\brief Redo the last undone editing operation
*/
void QDocument::redo()
{
	if ( m_impl )
		m_impl->m_commands.redo();
	
}

/*!
	\return The content of the document
	\param mode extra processing to perform on text
*/
QString QDocument::text(int mode) const
{
	QString s;
	
	if ( !m_impl || !m_impl->lineCount() )
		return s;
	
	//int line = 0;
	int prevIndent = 0, curIndent = 0, nextIndent = m_impl->at(0)->nextNonSpaceChar(0);
	
	if ( nextIndent < 0 )
		nextIndent = 0;
	
	QDocumentBuffer::iterator it = m_impl->m_buffer->begin();
	
	while ( !it.atEnd() )
	{
		QDocumentLineHandle *l = it.lineHandle();
		
		if ( !l )
			break;
		
		it.moveBy(1);
		
		prevIndent = curIndent;
		curIndent = nextIndent;
		
		nextIndent = it.atEnd() ? 0 : it.lineHandle()->nextNonSpaceChar(0);
		
		if ( nextIndent < 0 )
			nextIndent = 0;
		
		QString buf = l->text();
		int avgIndent = qMax(prevIndent, nextIndent);
		
		if ( (mode & RestoreTrailingIndent) && buf.isEmpty() && avgIndent )
		{
			buf = QString(avgIndent, '\t');
		} else if ( mode & RemoveTrailingWS ) {
			
			int len = 0, idx = buf.length();
			
			while ( --idx >= 0 )
			{
				if ( !buf.at(idx).isSpace() )
					break;
				
				++len;
			}
			
			++idx;
			
			if ( len && (idx || !(mode & PreserveIndent)) )
				buf.remove(idx, len);
		}
		
		s += buf;
		s += m_impl->m_lineEndingString;
	}
	
	//s.chop(m_impl->m_lineEndingString.count());
	return s;
}

/*!
	\return The content of the document
	\param removeTrailing whether to remove trailing whitespaces
	\param preserveIndent whether to keep trailing whitespaces when they are indent
*/
QString QDocument::text(bool removeTrailing, bool preserveIndent) const
{
	int mode = 0;
	
	if ( removeTrailing )
		mode |= RemoveTrailingWS;
	
	if ( preserveIndent )
		mode |= PreserveIndent;
	
	return text(mode);
}

void QDocument::load(const QString& filename, QTextCodec *c)
{
	m_impl->m_deleting = true;
	
	m_impl->m_marks.clear();
	m_impl->m_status.clear();
	m_impl->m_matches.clear();
	m_impl->m_largest.clear();
	m_impl->m_commands.clear();
	
	m_impl->m_deleting = false;
	
	m_impl->m_buffer->load(filename, c);
	m_impl->m_regions->reset(m_impl->lineCount());
	
	if ( lineEnding() == Conservative )
		setLineEnding(Conservative);
	
	//m_impl->setWidth();
	m_impl->setHeight();
	
	m_impl->m_lastModified = QFileInfo(filename).lastModified();
	
	// TODO : defered highligting?
	m_impl->emitContentsChange(0, m_impl->lineCount());
}

/*!
	\brief Set the content of the document
*/
void QDocument::setText(const QString& s)
{
	if ( !m_impl )
		return;
	
	m_impl->m_deleting = true;
	
	// TODO : smart clear (would need post load extra call) to avoid alloc/dealloc
	m_impl->m_buffer->clear();
	
	QDocumentCommand::discardHandlesFromDocument(this);
	
	m_impl->m_marks.clear();
	m_impl->m_status.clear();
	m_impl->m_matches.clear();
	m_impl->m_largest.clear();
	m_impl->m_commands.clear();
	
	m_impl->m_deleting = false;
	
	m_impl->m_buffer->setContent(s);
	m_impl->m_regions->reset(m_impl->lineCount());
	
	if ( lineEnding() == Conservative )
		setLineEnding(Conservative);
	
	m_impl->setWidth();
	m_impl->setHeight();
	
	m_impl->m_lastModified = QDateTime::currentDateTime();
	
	m_impl->emitContentsChange(0, m_impl->lineCount());
	emit lineCountChanged(lineCount());
}

/*!
	\brief Start a chunk loading
	
	It is possible to load document contents in one piece
	or by chunks. To achieve the later you have to proceed as follows :
	
	\code
	QDocument doc;
	doc.startChunkLoading();
	
	// fetch data and add it using doc.addChunk();
	
	doc.stopChunkLoading();
	\endcode
	
	\see addChunk(const QString&)
	\see stopChunkLoading()
*/
void QDocument::startChunkLoading()
{
	if ( !m_impl )
		return;
	
	m_impl->m_deleting = true;
	
	m_impl->m_buffer->clear();
	
	m_impl->m_marks.clear();
	m_impl->m_status.clear();
	m_impl->m_matches.clear();
	m_impl->m_largest.clear();
	m_impl->m_commands.clear();
	
	m_impl->m_deleting = false;
	
	m_impl->m_buffer->startChunkLoading();
}

/*!
	\brief Stop chunk loading
	
	\see startChunkLoading()
*/
void QDocument::stopChunkLoading()
{
	if ( !m_impl )
		return;
	
	m_impl->m_buffer->stopChunkLoading();
	
	if ( lineEnding() == Conservative )
		setLineEnding(Conservative);
	
	m_impl->m_regions->reset(m_impl->lineCount());
	
	m_impl->setWidth();
	m_impl->setHeight();
	
	m_impl->m_lastModified = QDateTime::currentDateTime();
	
	m_impl->emitContentsChange(0, m_impl->lineCount());
	emit lineCountChanged(lineCount());
}

/*!
	\return The document buffer where textual data is stored
*/
QDocumentBuffer* QDocument::buffer() const
{
	return m_impl->m_buffer;
}

/*!
	\return The format scheme used by the document
*/
QFormatScheme* QDocument::formatScheme() const
{
	return m_impl ? m_impl->m_formatScheme : 0;
}

/*!
	\brief Set the format scheme used by the document
*/
void QDocument::setFormatScheme(QFormatScheme *f)
{
	if ( m_impl )
		m_impl->setFormatScheme(f);
}

/*!
	\return the language definition set to the document
*/
QLanguageDefinition* QDocument::languageDefinition() const
{
	return m_impl ? m_impl->m_language : 0;
}

/*!
	\brief Set the language definition
*/
void QDocument::setLanguageDefinition(QLanguageDefinition *f)
{
	if ( m_impl )
		m_impl->m_language = f;
}

/*!
	\brief Update the formatting of the whole document
	This function is only useful when changing the language definition
	of a non-empty document. Make sure you do not call it more often
	than needed.
*/
void QDocument::highlight()
{
	if ( m_impl )
		m_impl->emitContentsChange(0, lines());
}

/*!
	\brief Add a chunk of text to the document
*/
void QDocument::addChunk(const QString& txt)
{
	if ( !m_impl )
		return;
	
	m_impl->m_buffer->addChunk(txt);
}

/*!
	\brief Print the content of the document
	\param pr printer to use
	
	\note the printer MUST be initialized (probably using a printing dialog)
*/
void QDocument::print(QPrinter *pr)
{
	QRect fit = pr->pageRect();
	
	if ( pr->printRange() == QPrinter::Selection )
	{
		qWarning("printing selection not implemented yet");
		return;
	}
	
	if ( fit.width() < width() )
	{
		// TODO: got to temporarily wrap text to fit page size
		
		qWarning("temporary wrapping not implementated yet");
	}
	
	const int lineCount = lines();
	const int linesPerPage = fit.height() / m_impl->m_lineSpacing;
	int pageCount = lineCount / linesPerPage;
	
	if ( lineCount % linesPerPage )
		++pageCount;
	
	//qDebug("%i lines per page -> %i pages", linesPerPage, pageCount);
	
	const int pageWidth = fit.width();
	const int pageHeight = linesPerPage * m_impl->m_lineSpacing;
	
	int firstPage = pr->fromPage(), lastPage = pr->toPage();
	
	if ( !lastPage )
		lastPage = pageCount - 1;
	
	QPainter p(pr);
	PaintContext cxt;
	cxt.xoffset = 0;
	cxt.yoffset = firstPage * pageHeight;
	cxt.width = pageWidth;
	cxt.height = pageHeight - m_impl->m_lineSpacing;
	cxt.palette = QApplication::palette();
	cxt.fillCursorRect = false;
	cxt.blinkingCursor = false;
	
	for ( int i = firstPage; i <= lastPage; ++i )
	{
		draw(&p, cxt);
		
		cxt.yoffset += pageHeight;
		
		if ( i != lastPage )
		{
			pr->newPage();
			p.translate(0, -pageHeight);
		}
	}
}

/*!
	\return The line ending policy of the document
	
	The line ending policy determines how line endings
	are used when saving the document (which includes
	fetching the document's text()).
	
	It can either be conservative (auto detect upon loading
	and do not modify when saving later on) or enforce
	a particular line ending (either local line ending
	or a specific value).
*/
QDocument::LineEnding QDocument::lineEnding() const
{
	return m_impl ? m_impl->m_lineEnding : Local;
}

/*!
	\return the line endings detected upon loading
	
	This should only ever take the the Window of Linux value
	if a document has been loaded. If no content has been
	loaded it will fall back to Local.
*/
QDocument::LineEnding QDocument::originalLineEnding() const
{
	return m_impl ? QDocument::LineEnding(m_impl->m_buffer->originalLineEnding()) : Local;
}

/*!
	\brief Set the line ending policy of the document
*/
void QDocument::setLineEnding(LineEnding le)
{
	if ( !m_impl )
		return;
	
	m_impl->m_lineEnding = le;
	QString& les = m_impl->m_lineEndingString;
	
	/*
		Notes about "local" line endings :
			* Dos & co : well defined CRLF
			* Unix & co : well defined LF
			* Mac : CR up until OS 9 LF afterwards (OSX is Unix-based)
		
		Guess about macros (to be confirmed/infirmed) :
			Q_OS_UNIX *should* be set under OSX
			Q_OS_MAC *might* be set under OSX
			
		=> order #ifdefs to try avoiding troubles
	*/
	#ifdef Q_OS_WIN
	const QString loc_les = "\r\n";
	#elif defined(Q_OS_UNIX)
	const QString loc_les = "\n";
	#elif defined(Q_OS_MAC)
	const QString loc_les = "\r";
	#else
	const QString loc_les = "\n";
	#endif
	
	int rle = le;
	
	if ( le == Conservative )
		rle = originalLineEnding();
	
	switch ( rle )
	{
		case Local :
			les = loc_les;
			break;
			
		case Unix :
			les = "\n";
			break;
			
		case OldMac :
			les = "\r";
			break;
			
		case Windows :
			les = "\r\n";
			break;
			
		default :
			les = "\n";
			break;
	}
	
	emit lineEndingChanged(le);
}

/*
QList<QDocumentMark*> QDocument::marks(int line) const
{
	QList<QDocumentMark*> l;
	
	if ( m_impl )
	{
		if ( line < 0 )
			l = m_impl->m_marks;
		else
			l = m_impl->m_marksLookup.value(m_impl->at(line));
	}
	
	return l;
}
*/

/*!
	\return the font used by ALL documents to render their content
	
	This font is also used to do calculations (such as converting
	(line, column) cursor position to (x, y) document position (or
	the inverse transformation))
	
	\note this limitation is historic and may disappear
	in future versions
*/
QFont QDocument::font()
{
	return *(QDocumentPrivate::m_font);
}

/*!
	\brief Set the font of ALL documents
	
	\note this limitation is historic and may disappear
	in future versions
*/
void QDocument::setFont(const QFont& f)
{
	QDocumentPrivate::setFont(f);
	//emit contentsChanged();
}

/*!
	\return The font metrics used by ALL documents
	
	\note this limitation is historic and may disappear
	in future versions
*/
const QFontMetrics& QDocument::fontMetrics()
{
	return *(QDocumentPrivate::m_fontMetrics);
}

/*!
	\return The default tab stop common to ALL documents
	
	\note this limitation is historic and may disappear
	in future versions
*/
int QDocument::tabStop()
{
	return QDocumentPrivate::m_defaultTabStop;
}

/*!
	\brief Set the default tab stop common to all documents
	
	\note this limitation is historic and may disappear
	in future versions
*/
void QDocument::setTabStop(int n)
{
	QDocumentPrivate::m_defaultTabStop = n;
	
	foreach ( QDocumentPrivate *d, QDocumentPrivate::m_documents )
	{
		d->m_tabStop = n;
		d->emitFormatsChanged();
	}
}

/*!
	\return the whitesapce display mode
*/
QDocument::WhiteSpaceMode QDocument::showSpaces()
{
	return QDocumentPrivate::m_showSpaces;
}

/*!
	\brief Set the whitespace display mode
*/
void QDocument::setShowSpaces(WhiteSpaceMode m)
{
	QDocumentPrivate::m_showSpaces = m;
	
	foreach ( QDocumentPrivate *d, QDocumentPrivate::m_documents )
		d->emitFormatsChanged();
		
}

/*!
	\brief Set the edit cursor
	
	Archaic concept designed for use in QEditor, no used anymore
*/
QDocumentCursor* QDocument::editCursor() const
{
	return m_impl ? m_impl->m_editCursor : 0;
}

/*!
	\brief Set the edit cursor
	
	\see editCursor()
*/
void QDocument::setEditCursor(QDocumentCursor *c)
{
	if ( m_impl )
		m_impl->m_editCursor = c;
		
}

/*!
	\return the width of the document, in pixels
	
	The width of the document is that of longest text line.
*/
int QDocument::width() const
{
	return m_impl ? m_impl->m_width : 0;
}

/*!
	\return the height of the document, in pixels
*/
int QDocument::height() const
{
	return m_impl ? m_impl->m_height : 0;
}

/*!
	\return The width constraint imposed on that document
	
	Setting a width constraint on a document achieves line
	wrapping.
*/
int QDocument::widthConstraint() const
{
	return (m_impl && m_impl->m_constrained) ? m_impl->m_width : 0;
}

/*!
	\return the number of text lines in the document
	
	\deprecated Use lineCount() instead
*/
int QDocument::lines() const
{
	return m_impl ? m_impl->lineCount() : 0;
}

/*!
	\return the number of text lines in the document
	
	The number of visual lines may differ from that of text
	lines as soon as line wrapping and/or folding are enabled.
*/
int QDocument::lineCount() const
{
	return m_impl ? m_impl->lineCount() : 0;
}

/*!
	\return the number of visual lines in the document
*/
int QDocument::visualLineCount() const
{
	return m_impl ? m_impl->visualLine(m_impl->lineCount() - 1) : 0;
}

/*!
	\return the number of visual lines in the document
	
	\deprecated Use visualLineCount() instead
*/
int QDocument::visualLines() const
{
	return m_impl ? m_impl->visualLine(m_impl->lineCount() - 1) : 0;
}

/*!
	\brief Convert a text (logical) line number int a visual line number
	
	\note this is not a 1:1 mapping as logical lines can span over several visual lines
*/
int QDocument::visualLineNumber(int textLineNumber) const
{
	return m_impl ? m_impl->visualLine(textLineNumber) : -1;
}

/*!
	\brief Convert a visual line number int a text (logical) line number
	
	\note this is not a 1:1 mapping as logical lines can span over several visual lines
*/
int QDocument::textLineNumber(int visualLineNumber) const
{
	return m_impl ? m_impl->textLine(visualLineNumber) : -1;
}

/*!
	\brief Clear the width constraint, if any
*/
void QDocument::clearWidthConstraint()
{
	if ( m_impl )
		m_impl->setWidth(0);
}

/*!
	\brief Set a new width constraint
	\param width maximum width to allow
	
	Passing a value inferior (or equal) to zero clear the width constraint, if any.
*/
void QDocument::setWidthConstraint(int width)
{
	if ( m_impl )
		m_impl->setWidth(qMax(0, width));
}

/*!
	\return the line object at a given line number
	\param line Text line to acces
*/
QDocumentLine QDocument::line(int line) const
{
	return QDocumentLine(this, m_impl->at(line));
}

/*!
	\return the line number corresponding to a given document y coordinate
	\param ypos Y document coordinate of the target
	\param wrap if not null, will be set to the wrap offset (position of the
	visual line among the sublines of the wrapped text line).
	
*/
int QDocument::lineNumber(int ypos, int *wrap) const
{
	int ln = ypos / m_impl->m_lineSpacing;
	
	return m_impl->textLine(ln, wrap);
}

/*!
	\return A cursor operating on the document, placed at a given position
	\param line target line number (text line)
	\param column target text column
*/
QDocumentCursor QDocument::cursor(int line, int column) const
{
	return QDocumentCursor(const_cast<QDocument*>(this), line, column);
}

/*!
	\return The Y document coordinate of a given line
	\param ln textual line number
*/
int QDocument::y(int ln) const
{
	if ( !m_impl )
		return -1;
	
	return m_impl->m_lineSpacing * m_impl->visualLine(ln);
}

/*!
	\overload
	
	\return The Y document coordinate of a given line
	\param l line object
	
	\note Can be significantly slower than the line number based version.
*/
int QDocument::y(const QDocumentLine& l) const
{
	return y(l.lineNumber());
}

/*!
	\return the rectangle (in document position) occupied by the line
	\param line textual line number
	
	\note the width of the returned rectangle is the DOCUMENT's width
*/
QRect QDocument::lineRect(int line) const
{
	const int yoff = y(line);
	
	return (yoff != -1) ? QRect(0, yoff, width(), this->line(line).lineSpan() * m_impl->m_lineSpacing) : QRect();
}

/*!
	\overload
	\return the rectangle (in document position) occupied by the line
	
	\note the width of the returned rectangle is the DOCUMENT's width
	\note Significantly slower than the line number based version.
*/
QRect QDocument::lineRect(const QDocumentLine& l) const
{
	//return lineRect(l.lineNumber());
	const int yoff = y(l);
	
	return (yoff != -1) ? QRect(0, yoff, width(), m_impl->m_lineSpacing) : QRect();
}

/*!
	\return the line at a given document position
*/
QDocumentLine QDocument::lineAt(const QPoint& p) const
{
	if ( !m_impl )
		return QDocumentLine();
	
	return line(lineNumber(p.y()));
}

/*!
	\brief Convert a document (or viewport) (x, y) position to a (line, column) cursor position
	\param p document position
	\param line where the line number will be stored
	\param column where the column (text position within line) will be stored
*/
void QDocument::cursorForDocumentPosition(const QPoint& p, int& line, int& column) const
{
	if ( !m_impl )
		return;
	
	int wrap = 0;
	line = lineNumber(p.y(), &wrap);
	QDocumentLine l = this->line(line);
	
	if ( !l.isValid() )
		return;
	
	//qDebug("%i %i", line, wrap);
	column = l.documentOffsetToCursor(p.x(), wrap * QDocumentPrivate::m_lineSpacing);
	
	//qDebug("(%i, %i) -> (%i [+%i], %i)", p.x(), p.y(), line, wrap, column);
}

/*!
	\return The cursor nearest to a document (x, y) position
*/
QDocumentCursor QDocument::cursorAt(const QPoint& p) const
{
	int ln = -1, col = -1;
	
	cursorForDocumentPosition(p, ln, col);
	
	return QDocumentCursor(const_cast<QDocument*>(this), ln, col);
}

/*!
	\brief Draw the contents of the document
	\param p painter to use
	\param cxt paint context (specifies what part of the document to draw, among other things)
*/
void QDocument::draw(QPainter *p, PaintContext& cxt)
{
	m_impl->draw(p, cxt);
}

/*!
	\brief Execute a document command (editing operation)
*/
void QDocument::execute(QDocumentCommand *cmd)
{
	if ( m_impl && cmd )
		m_impl->execute(cmd);
		
}

/*!
	\return The default line ending policy
*/
QDocument::LineEnding QDocument::defaultLineEnding()
{
	return QDocumentPrivate::m_defaultLineEnding;
}

/*!
	\brief Set the default line ending policy
	
	\note The line ending policy of existing documents is changed accordingly
*/
void QDocument::setDefaultLineEnding(QDocument::LineEnding le)
{
	QDocumentPrivate::m_defaultLineEnding = le;
	
	foreach ( QDocumentPrivate *d, QDocumentPrivate::m_documents )
	{
		d->m_doc->setLineEnding(le);
	}
}

/*!
	\return The default format scheme used to draw document contents
*/
QFormatScheme* QDocument::defaultFormatScheme()
{
	return QDocumentPrivate::m_defaultFormatScheme;
}

/*!
	\brief Set the default format scheme
	
	\note Existing documents using the default format scheme will see their format scheme changed
*/
void QDocument::setDefaultFormatScheme(QFormatScheme *f)
{
	foreach ( QDocumentPrivate *d, QDocumentPrivate::m_documents )
	{
		if ( d->m_formatScheme == QDocumentPrivate::m_defaultFormatScheme )
			d->setFormatScheme(f);
	}
	
	QDocumentPrivate::m_defaultFormatScheme = f;
}

/*!
	\deprecated
	\brief Compatibility alias for defaultFormatScheme()
*/
QFormatScheme* QDocument::formatFactory()
{
	return defaultFormatScheme();
}

/*!
	\deprecated
	\brief Compatibility alias for setDefaultFormatScheme()
*/
void QDocument::setFormatFactory(QFormatScheme *f)
{
	setDefaultFormatScheme(f);
}

/*!
	\brief Begin a macro
	
	Macro in QDocument map directly to macro in the underlying undo stack
*/
void QDocument::beginMacro()
{
	if ( m_impl )
		m_impl->beginChangeBlock();
		
}

/*!
	\brief End a macro
*/
void QDocument::endMacro()
{
	if ( m_impl )
		m_impl->endChangeBlock();
		
}

/*!
	\brief Get an available group id for matches
*/
int QDocument::getNextGroupId()
{
	return m_impl ? m_impl->getNextGroupId() : -1;
}

void QDocument::releaseGroupId(int groupId)
{
	if ( m_impl )
		m_impl->releaseGroupId(groupId);
}

/*!
	\brief Clear matches
*/
void QDocument::clearMatches(int gid)
{
	if ( m_impl )
	{
		m_impl->clearMatches(gid);
	}
}

/*!
	\brief Highlight the matched sequences
	
	\note Both position are BEFORE the matched characters (cursor position).
*/
void QDocument::addMatch(int gid, int line, int pos, int len, int format)
{
	if ( m_impl )
	{
		m_impl->addMatch(gid, line, pos, len, format);
	}
}

/*!
	\
*/
void QDocument::flushMatches(int gid)
{
	if ( m_impl )
	{
		m_impl->flushMatches(gid);
	}
}

/*!
	\return Whether the document is in a clean state
	
	This is undo stak terminology. A clean document is one
	whose undo stack is at the same index it was upon last
	save/load. In other words : clean = !modified
*/
bool QDocument::isClean() const
{
	return m_impl ? m_impl->m_commands.isClean() : true;
}

/*!
	\brief Set the document to clean state
	
	This method does not go back to clean state but tell
	the stack that the current state is to be considered
	as the clean state.
*/
void QDocument::setClean()
{
	if ( m_impl )
	{
		m_impl->m_commands.setClean();
		//m_impl->m_status.clear();
		
		QHash<QDocumentLineHandle*, QPair<int, int> >::iterator it = m_impl->m_status.begin();
		
		while ( it != m_impl->m_status.end() )
		{
			it->second = it->first;
			++it;
		}
	}
}

/*!
	\return Whether a given line has been modified since last save/load
*/
bool QDocument::isLineModified(const QDocumentLine& l) const
{
	if ( !m_impl )
		return false;
	
	QHash<QDocumentLineHandle*, QPair<int, int> >::const_iterator it = m_impl->m_status.constFind(l.handle());
	
	//if ( it != m_impl->m_status.constEnd() )
	//	qDebug("%i vs %i", it->first, it->second);
	
	return it != m_impl->m_status.constEnd() ? it->first != it->second : false;
}

/*!
	\return Whether a given line has been modified since last load
*/
bool QDocument::hasLineEverBeenModified(const QDocumentLine& l) const
{
	return m_impl ? m_impl->m_status.contains(l.handle()) : false;
}

/*!
	\return the maximum number of marks on a single line
	
	This is meant for the line mark panel to smartly adapt its size.
*/
int QDocument::maxMarksPerLine() const
{
	return m_impl ? m_impl->maxMarksPerLine() : 0;
}

/*!
	\brief Find the next mark of a given type
	\return the line on which a mark was found
	\param id mark identifier to find
	\param from line from which to start search
	\param until line until which to search
	
	\a from and \a until can be negatives, in which case they
	indicate positions from the end of the document (i.e -1 is
	last line, -2 the line before last line and so on).
	
	If \a until is inferior to \a from and no matching mark
	is found in the [from, end] range then the [0, until]
	range will also be searched.
*/
int QDocument::findNextMark(int id, int from, int until) const
{
	return m_impl ? m_impl->findNextMark(id, from, until) : -1;
}

/*!
	\brief Find the previous mark of a given type
	\return the line on which a mark was found
	\param id mark identifier to find
	\param from line from which to start search
	\param until line until which to search
	
	\a from and \a until can be negatives, in which case they
	indicate positions from the end of the document (i.e -1 is
	last line, -2 the line before last line and so on).
	
	If \a until is superior to \a from and no matching mark
	is found in the [0, from] range then the [until, end]
	range will also be searched (both range being searched backward,
	of course).
*/
int QDocument::findPreviousMark(int id, int from, int until) const
{
	return m_impl ? m_impl->findPreviousMark(id, from, until) : -1;
}

void QDocument::addMark(QDocumentMark *m)
{
	m_impl->addMark(m);
}

void QDocument::removeMark(QDocumentMark *m)
{
	m_impl->removeMark(m);
}

QList<QDocumentMark*> QDocument::marks() const
{
	return m_impl->m_marks;
}

QList<QDocumentMark*> QDocument::findMarks(int line) const
{
	return findMarks(QDocumentCursor(const_cast<QDocument*>(this), line), QList<int>());
}

QList<QDocumentMark*> QDocument::findMarks(const QList<int>& ids) const
{
	return findMarks(QDocumentCursor(), ids);
}

QList<QDocumentMark*> QDocument::findMarks(int line, int id) const
{
	return findMarks(QDocumentCursor(const_cast<QDocument*>(this), line), QList<int>() << id);
}

QList<QDocumentMark*> QDocument::findMarks(const QDocumentCursor& c, int id) const
{
	return findMarks(c, QList<int>() << id);
}

QList<QDocumentMark*> QDocument::findMarks(const QDocumentCursor& c, const QList<int>& ids) const
{
	return m_impl->findMarks(c.handle(), ids);
}

/*!
	\return the date/time of the last modification of the document
	
	If the document has not been modified since its load the date/time
	of last modification (as reported by QFileInfo) will be returned.
*/
QDateTime QDocument::lastModified() const
{
	return m_impl ? m_impl->m_lastModified : QDateTime();
}

/*!
	\brief set the date/time of the last modification of the document
	
	You should not need to use that EVER. It is only provided for use
	in QEditor (and possibly in some panels).
*/
void QDocument::setLastModified(const QDateTime& d)
{
	if ( m_impl )
		m_impl->m_lastModified = d;
}

/////////////////////////
//	QDocumentLineHandle
/////////////////////////
//static quint32 q_line_counter = 0;

/*!
	\class QDocumentLineHandle
	\brief Private implementation of a document line
*/

/*!
	\
*/
QDocumentLineHandle::QDocumentLineHandle()
 : m_state(QDocumentLine::TextChanged | QDocumentLine::FormattingChanged)
{
	m_render.data = 0;
}

/*!
	\
*/
QDocumentLineHandle::QDocumentLineHandle(const QString& s)
 : m_state(QDocumentLine::TextChanged | QDocumentLine::FormattingChanged)
 , m_text(s)
{
	m_render.data = 0;
}

/*!
	\
*/
QDocumentLineHandle::QDocumentLineHandle(const QChar *d, int size)
 : m_state(QDocumentLine::TextChanged | QDocumentLine::FormattingChanged)
 , m_text(d, size)
{
	m_render.data = 0;
}

QDocumentLineHandle::~QDocumentLineHandle()
{
	
}

int QDocumentLineHandle::count() const
{
	return m_text.count();
}

int QDocumentLineHandle::length() const
{
	return m_text.length();
}

QString QDocumentLineHandle::text() const
{
	return m_text;
}

int QDocumentLineHandle::nextNonSpaceChar(uint pos) const
{
	const int len = m_text.length();
	const QChar *unicode = m_text.unicode();
	
	for ( int i = pos; i < len; ++i )
	{
		if ( !unicode[i].isSpace() )
			return i;
	}
	
	return -1;
}

int QDocumentLineHandle::previousNonSpaceChar(int pos) const
{
	const int len = m_text.length();
	const QChar *unicode = m_text.unicode();
	
	if ( pos < 0 )
		pos = 0;
		
	if ( pos >= len )
		pos = len - 1;
		
	for ( int i = pos; i >= 0; --i )
	{
		if ( !unicode[i].isSpace() )
			return i;
	}
	
	return -1;
}

bool QDocumentLineHandle::hasFlag(int s) const
{
	return m_state & s;
}

void QDocumentLineHandle::setFlag(int s, bool y)
{
	if ( y )
	{
		//if ( s == QDocumentLine::FormattingChanged )
		//	qDebug("fmtchg");
		
		m_state |= s;
	} else {
		m_state &= ~s;
	}
}

void QDocumentLineHandle::clearOverlays()
{
	if ( m_tokens.isEmpty() )
		return;
	
	int overlayIdx = m_tokens.first();
	
	if ( overlayIdx >= m_tokens.count() )
		return;
	
	m_tokens.resize(overlayIdx);
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::addOverlay(const QFormatRange& over)
{
	if ( m_tokens.isEmpty() )
	{
		m_tokens << 1;
	}
	
	m_tokens << over.offset << over.length << over.format;
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::removeOverlay(const QFormatRange& over)
{
	if ( m_tokens.isEmpty() )
		return;
	
	int overlayIdx = m_tokens.first();
	
	if ( overlayIdx >= m_tokens.count() )
		return;
	
	while ( overlayIdx + 2 < m_tokens.count() )
	{
		if ( m_tokens.at(overlayIdx + 2) == over.format )
		{
			// TODO : loose matching to take shifting into account?
			if ( m_tokens.at(overlayIdx) == over.offset && m_tokens.at(overlayIdx + 1) == over.length )
			{
				m_tokens.remove(overlayIdx);
				m_tokens.remove(overlayIdx);
				m_tokens.remove(overlayIdx);
				setFlag(QDocumentLine::FormattingChanged, true);
				continue;
			}
		}
		
		overlayIdx += 3;
	}
}

void QDocumentLineHandle::shiftOverlays(int position, int offset)
{
	if ( m_tokens.isEmpty() )
		return;
	
	int overlayIdx = m_tokens.first();
	
	if ( overlayIdx >= m_tokens.count() )
		return;
	
	if ( offset > 0 )
	{
		while ( overlayIdx + 2 < m_tokens.count() )
		{
			int off = m_tokens.at(overlayIdx);
			int len = m_tokens.at(overlayIdx + 1);
			
			if ( off >= position )
			{
				m_tokens[overlayIdx] += offset;
			} else if ( off + len > position ) {
				m_tokens.remove(overlayIdx);
				m_tokens.remove(overlayIdx);
				m_tokens.remove(overlayIdx);
				continue;
			}
			
			overlayIdx += 3;
		}
	} else if ( offset < 0 ) {
		const int max = position - offset;
		
		while ( overlayIdx + 2 < m_tokens.count() )
		{
			int off = m_tokens.at(overlayIdx);
			int len = m_tokens.at(overlayIdx + 1);
			
			if ( off >= max )
			{
				m_tokens[overlayIdx] += offset;
			} else if ( off + len >= position ) {
				m_tokens.remove(overlayIdx);
				m_tokens.remove(overlayIdx);
				m_tokens.remove(overlayIdx);
				continue;
			}
			
			overlayIdx += 3;
		}
	}
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::clearTokens()
{
	int overlayIdx = m_tokens.count() ? m_tokens.first() : 0;
	
	if ( overlayIdx > 1 && overlayIdx < m_tokens.count() )
	{
		m_tokens.remove(1, overlayIdx - 1);
	} else {
		m_tokens.clear();
		//m_tokens << 1;
	}
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::addToken(int start, int length, int format)
{
	int overlayIdx = m_tokens.count() ? m_tokens.first() : 0;
	
	//qDebug("adding token to ((%i))", overlayIdx);
	
	if ( overlayIdx < 3 )
	{
		m_tokens.clear();
		m_tokens << 4; // 1 + 3
		m_tokens << start << length << format;
	} else {
		m_tokens[0] += 3;
		m_tokens.insert(overlayIdx, start);
		m_tokens.insert(++overlayIdx, length);
		m_tokens.insert(++overlayIdx, format);
		//<< start << length << format;
	}
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::setTokens(const QVector<int>& tokens)
{
	QVector<int> remaining;
	
	if ( m_tokens.count() )
	{
		int overlayIdx = m_tokens.first();
		int count = m_tokens.count() - overlayIdx;
		
		if ( count )
		{
			remaining.reserve(count);
			memcpy(remaining.data(), m_tokens.constData() + overlayIdx, count);
		}
		
		m_tokens.clear();
	}
	
	m_tokens << tokens.count() + 1;
	m_tokens << tokens;
	
	if ( remaining.count() )
		m_tokens << remaining;
	
	//setFlag(QDocumentLine::LayoutDirty, true);
	setFlag(QDocumentLine::FormattingChanged, true);
	//applyOverlays();
}

void QDocumentLineHandle::clearContexts()
{
	m_contexts.clear();
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::addContext(int start, int length, int format)
{
	m_contexts << start << length << format;
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

void QDocumentLineHandle::setContexts(const QVector<int>& contexts)
{
	m_contexts = contexts;
	
	setFlag(QDocumentLine::FormattingChanged, true);
}

QList<QTextLayout::FormatRange> QDocumentLineHandle::decorations() const
{
	// turning format data into ranges that QTextLayout can understand...
	QList<QTextLayout::FormatRange> m_ranges;
	
	int i = 1;
	
	while ( i + 2 < m_tokens.count() )
	{
		QTextLayout::FormatRange r;
		
		r.start = m_tokens.at(i);
		r.length = m_tokens.at(++i);
		// FIXME : pass a pointer to the scheme
		//r.format = m_doc->formatScheme()->format(m_tokens.at(++i)).toTextCharFormat();
		
		m_ranges << r;
	}
	
	return m_ranges;
}

void QDocumentLineHandle::applyOverlays()
{
	
}

//////////////////


/////////////////////////
//	QDocumentCursorHandle
/////////////////////////
QDocumentCursorHandle::QDocumentCursorHandle(QDocument *d, int line)
 :	m_flags(ColumnMemory), m_doc(d),
	#if QT_VERSION >= 0x040400
	m_ref(0),
	#endif
	m_begOffset(0), m_endOffset(0), m_max(0), m_begLine(line), m_endLine(-1)
{
	#if QT_VERSION < 0x040400
	m_ref.init(0);
	#endif
	
	//m_blocks.push(0);
	//qDebug("Cursor handle created : 0x%x", this);
}

QDocumentCursorHandle::~QDocumentCursorHandle()
{
	//qDebug("Cursor handle deleted : 0x%x", this);
	Q_ASSERT(!m_ref);
	
	QDocumentCommand::disableAutoUpdate(this);
}

void QDocumentCursorHandle::copy(const QDocumentCursorHandle *c)
{
	if ( !c )
		return;
		
	m_begLine = c->m_begLine;
	m_begOffset = c->m_begOffset;
	m_endLine = c->m_endLine;
	m_endOffset = c->m_endOffset;
	m_flags = c->m_flags;
	m_max = c->m_max;
}

QDocument* QDocumentCursorHandle::document() const
{
	return m_doc;
}

QDocumentCursorHandle* QDocumentCursorHandle::clone() const
{
	QDocumentCursorHandle *c = new QDocumentCursorHandle(m_doc);
	c->copy(this);
	
	c->setAutoUpdated(isAutoUpdated());
	
	return c;
}

bool QDocumentCursorHandle::atEnd() const
{
	if ( !m_doc )
		return true;
		
	bool atLineEnd;
	QDocumentLine l = m_doc->line(m_begLine);
	
	//qDebug("Cursor handle : 0x%x->atEnd() => 0x%x", this, m_begLine.handle());
	
	if ( l.isValid() )
	{
		atLineEnd = m_begOffset == l.length();
		l = m_doc->line(m_begLine + 1);
	} else {
		//qWarning("Invalid cursor...");
		return true;
	}
	
	return l.isNull() && atLineEnd;
}

bool QDocumentCursorHandle::atStart() const
{
	if ( !m_doc )
		return true;
		
	QDocumentLine l = m_doc->line(m_begLine - 1);
	
	return l.isNull() && !m_begOffset;
}

bool QDocumentCursorHandle::atBlockEnd() const
{
	return atLineEnd();
}

bool QDocumentCursorHandle::atBlockStart() const
{
	return atLineStart();
}

bool QDocumentCursorHandle::atLineEnd() const
{
	if ( !m_doc )
		return true;
		
	QDocumentLine l = m_doc->line(m_begLine);
	
	return l.isValid() ? l.length() == m_begOffset : false;
}

bool QDocumentCursorHandle::atLineStart() const
{
	if ( !m_doc )
		return true;
		
	QDocumentLine l = m_doc->line(m_begLine);
	
	return l.isValid() ? !m_begOffset : false;
}

bool QDocumentCursorHandle::hasSelection() const
{
	if ( !m_doc )
		return false;
		
	QDocumentLine l1 = m_doc->line(m_begLine), l2 = m_doc->line(m_endLine);
	
	return l1.isValid() && l2.isValid();
}

bool QDocumentCursorHandle::isSilent() const
{
	return hasFlag(Silent);
}

void QDocumentCursorHandle::setSilent(bool y)
{
	if ( y )
		setFlag(Silent);
	else
		clearFlag(Silent);
}

bool QDocumentCursorHandle::isAutoUpdated() const
{
	return QDocumentCommand::isAutoUpdated(this);
}

void QDocumentCursorHandle::setAutoUpdated(bool y)
{
	if ( y )
		QDocumentCommand::enableAutoUpdate(this);
	else
		QDocumentCommand::disableAutoUpdate(this);
}

QDocumentLine QDocumentCursorHandle::line() const
{
	if ( !m_doc )
		return QDocumentLine();
		
	return m_doc->line(m_begLine);
}

QDocumentLine QDocumentCursorHandle::anchorLine() const
{
	if ( !m_doc )
		return QDocumentLine();
		
	return m_endLine != -1 ? m_doc->line(m_endLine) : line();
}

QList<QDocumentLineHandle*> QDocumentCursorHandle::selectionSpan() const
{
	QList<QDocumentLineHandle*> l;
	
	if ( m_endLine != -1 )
	{
		int first = qMin(m_begLine, m_endLine), last = qMax(m_begLine, m_endLine);
		
		for ( int i = first; i <= last; ++i )
			l << m_doc->impl()->at(i);
	} else {
		l << m_doc->impl()->at(m_begLine);
	}
	
	return l;
}

int QDocumentCursorHandle::lineNumber() const
{
	return m_begLine;
}

int QDocumentCursorHandle::anchorLineNumber() const
{
	return m_endLine != -1 ? m_endLine : m_begLine;
}

int QDocumentCursorHandle::anchorColumnNumber() const
{
	if ( !m_doc )
		return -1;
		
	return m_doc->line(m_endLine).isValid() ? m_endOffset : m_begOffset;
}

int QDocumentCursorHandle::visualColumnNumber() const
{
	return QDocument::screenLength(
						line().text().constData(),
						m_begOffset,
						QDocument::tabStop()
					);
					
}

int QDocumentCursorHandle::columnNumber() const
{
	return m_begOffset;
}

void QDocumentCursorHandle::setColumnNumber(int c, int m)
{
	if ( !m_doc )
		return;
		
	QDocumentLine l1 = m_doc->line(m_begLine), l2 = m_doc->line(m_endLine);
	
	if ( m & QDocumentCursor::KeepAnchor )
	{
		if ( l2.isNull() )
		{
			m_endLine = m_begLine;
			m_endOffset = m_begOffset;
		}
		
		m_begOffset = c; //qBound(0, c, l1.length());
	} else {
		m_endLine = -1;
		m_endOffset = 0;
		m_begOffset = c; //qBound(0, c, l1.length());
	}
	
	refreshColumnMemory();
}

QPoint QDocumentCursorHandle::documentPosition() const
{
	if ( !m_doc )
		return QPoint();
	
	return QPoint(0, m_doc->y(m_begLine)) + m_doc->line(m_begLine).cursorToDocumentOffset(m_begOffset);
}

QPoint QDocumentCursorHandle::anchorDocumentPosition() const
{
	if ( !m_doc )
		return QPoint();
	
	if ( m_endLine < 0 || m_endOffset < 0 )
		return documentPosition();
	
	return QPoint(0, m_doc->y(m_endLine)) + m_doc->line(m_endLine).cursorToDocumentOffset(m_endOffset);
}

QPolygon QDocumentCursorHandle::documentRegion() const
{
	QPolygon poly;
	QPoint p = documentPosition(), ap = anchorDocumentPosition();
	
	int w = m_doc->width();
	const int lm = m_doc->impl()->m_leftMargin;
	const int ls = m_doc->impl()->m_lineSpacing - 1;
	
	if ( p == ap )
	{
		poly
			<< p
			<< QPoint(p.x() + 1, p.y())
			<< QPoint(p.x() + 1, p.y() + ls)
			<< QPoint(p.x(), p.y() + ls);
	} else if ( p.y() == ap.y() ) {
		poly
			<< p
			<< ap
			<< QPoint(ap.x(), ap.y() + ls)
			<< QPoint(p.x(), p.y() + ls);
	} else if ( p.y() < ap.y() ) {
		poly
			<< p
			<< QPoint(w, p.y());
		
		if ( ap.x() < w )
			poly << QPoint(w, ap.y()) << ap;
		
		poly
			<< QPoint(ap.x(), ap.y() + ls)
			<< QPoint(lm, ap.y() + ls)
			<< QPoint(lm, p.y() + ls);
		
		if ( p.x() > lm )
			poly << QPoint(p.x(), p.y() + ls);
	} else {
		poly
			<< ap
			<< QPoint(w, ap.y());
		
		if ( p.x() < w )
			poly << QPoint(w, p.y()) << p;
		
		poly
			<< QPoint(p.x(), p.y() + ls)
			<< QPoint(lm, p.y() + ls)
			<< QPoint(lm, ap.y() + ls);
		
		if ( ap.x() > lm )
			poly << QPoint(ap.x(), ap.y() + ls);
	}
	
	return poly;
}

int QDocumentCursorHandle::position() const
{
	// TODO : remove
	return -1;
	/*
	if ( !m_doc )
		return -1;
		
	int pos = m_doc->line(m_begLine).position();
	
	return (pos != -1) ? pos + m_begOffset : pos;
	*/
}

void QDocumentCursorHandle::shift(int offset)
{
	if ( !m_doc )
		return;
		
	QDocumentLine l1 = m_doc->line(m_begLine), l2 = m_doc->line(m_endLine);
	
	if ( l1.isValid() )
		m_begOffset = qBound(0, m_begOffset + offset, l1.length());
		
	if ( l2.isValid() )
		m_endOffset = qBound(0, m_endOffset + offset, l2.length());
}

void QDocumentCursorHandle::refreshColumnMemory()
{
	//m_max = m_doc->line(line).cursorToX(offset);
	m_max = hasFlag(ColumnMemory) ? m_doc->line(m_begLine).cursorToDocumentOffset(m_begOffset).x() : 0;
}

bool QDocumentCursorHandle::hasColumnMemory() const
{
	return hasFlag(ColumnMemory);
}

void QDocumentCursorHandle::setColumnMemory(bool y)
{
	if ( y )
		setFlag(ColumnMemory);
	else
		clearFlag(ColumnMemory);
}

void QDocumentCursorHandle::setPosition(int pos, int m)
{
	Q_UNUSED(pos)
	Q_UNUSED(m)
	
	qWarning("Set position to cursor using character index : forbidden...");
	/*
	if ( m == QDocumentCursor::MoveAnchor )
	{
		m_begLine = m_doc->findLine(pos);
		m_begOffset = (m_begLine.isValid() ? pos : 0);
		
		m_endLine = QDocumentLine();
		m_endOffset = 0;
		
		m_max = m_begLine.cursorToX(m_begOffset);
	} else {
		m_endLine = m_doc->findLine(pos);
		m_endOffset = (m_begLine.isValid() ? pos : 0);
		
		//m_max = 0;
	}
	*/
}

bool QDocumentCursorHandle::movePosition(int count, int op, int m)
{
	if ( !m_doc )
		return false;
	
	QDocumentLine l, l1 = m_doc->line(m_begLine), l2 = m_doc->line(m_endLine);
	
	int &line = m_begLine;
	int &offset = m_begOffset;
	static QRegExp wordStart("\\b\\w+$"), wordEnd("^\\w+\\b");
	
	if ( !(m & QDocumentCursor::KeepAnchor) )
	{
		m_endLine = -1;
		m_endOffset = 0;
	} else if ( !l2.isValid() ) {
		m_endLine = m_begLine;
		m_endOffset = m_begOffset;
	}
	
	int beg = 0, end = m_doc->lines();
	
	switch ( op )
	{
		case QDocumentCursor::Left :
		{
			if ( atStart() )
				return false;
				
			int remaining = offset;
			
			do
			{
				if ( remaining >= count )
				{
					offset = remaining - count;
					break;
				} else if ( line == beg ) {
					offset = 0;
					break;
				}
				
				do
				{
					--line;
				} while ( (line > beg) && m_doc->line(line).hasFlag(QDocumentLine::Hidden) );
				
				//*line = *it;
				
				count -= remaining + 1; // jumping a line is one char
				offset = remaining = m_doc->line(line).length();
			} while ( count && remaining );
			
			refreshColumnMemory();
			
			break;
		}
		
		case QDocumentCursor::Right :
		{
			if ( atEnd() )
				return false;
				
			int remaining = m_doc->line(line).length() - offset;
			
			do
			{
				if ( remaining >= count )
				{
					offset += count;
					break;
				} else if ( (line + 1) == end ) {
					offset = remaining;
					break;
				}
				
				do
				{
					++line;
				} while ( ((line+1) < end) && m_doc->line(line).hasFlag(QDocumentLine::Hidden) );
				
				//*line = *it;
				
				offset = 0;
				count -= remaining + 1;
				remaining = m_doc->line(line).length();
			} while ( count && remaining );
			
			refreshColumnMemory();
			
			break;
		}
		
		case QDocumentCursor::Up :
		{
			if ( atStart() )
				return false;
			
			//qDebug("%i, %i  : up", line, offset);
			
			if ( m & QDocumentCursor::ThroughWrap )
			{
				QPoint p = documentPosition();
				
				if ( hasColumnMemory() )
					p.rx() = qMax(p.x(), m_max);
				
				p.ry() -= QDocumentPrivate::m_lineSpacing * count;
				
				if ( p.y() >= 0 )
				{
					m_doc->cursorForDocumentPosition(p, line, offset);
				} else {
					line = 0;
					offset = 0;
				}
				
				return true;
			}
			
			while ( count && (line > beg) )
			{
				--line;
				
				if ( !m_doc->line(line).hasFlag(QDocumentLine::Hidden) )
					--count;
					
			}
			
			//*line = QDocumentLine(*it);
			//*offset = line->xToCursor(qMin(line->cursorToX(*offset), m_max), 0);
			l = m_doc->line(line);
			
			if ( count )
				offset = 0;
			else if ( m == QDocumentCursor::MoveAnchor ) {
				
				// FIXME : port to doc offset
				/*
				offset = l.xToCursor(
						qMax(
							l.cursorToX(
								qMin(
									offset,
									l.length()
								)
							),
							m_max
						)
					);
				*/
			} else
				offset = qMin(l.length(), offset);
				
			break;
		}
		
		case QDocumentCursor::Down :
		{
			if ( atEnd() )
				return false;
				
			if ( m & QDocumentCursor::ThroughWrap )
			{
				QPoint p = documentPosition();
				
				if ( hasColumnMemory() )
					p.rx() = qMax(p.x(), m_max);
					
				p.ry() += QDocumentPrivate::m_lineSpacing * count;
				p.ry() = qMin(p.ry(), m_doc->height() - QDocumentPrivate::m_lineSpacing / 2);
				
				int oldLine = line, oldCol = offset;
				m_doc->cursorForDocumentPosition(p, line, offset);
				if ( oldLine == line && oldCol == offset )
					offset = m_doc->line(line).length();
				return true;
			}
			
			while ( count && ((line + 1) < end) )
			{
				++line;
				
				if ( !m_doc->line(line).hasFlag(QDocumentLine::Hidden) )
					--count;
					
			}
			
			//*line = QDocumentLine(*it);
			l = m_doc->line(line);
			
			if ( count )
				offset = l.length();
			else if ( m == QDocumentCursor::MoveAnchor ) {
				// FIXME : port to doc offset
				/*
				offset = l.xToCursor(
						qMax(
							l.cursorToX(
								qMin(
									offset,
									l.length()
								)
							),
							m_max
						)
					);
				*/
			} else
				offset = qMin(l.length(), offset);
				
			break;
		}
		
		case QDocumentCursor::Start :
			if ( atStart() )
				return false;
				
			m_max = offset = 0;
			line = 0; //m_doc->line(0);
			break;
			
		case QDocumentCursor::End :
			if ( atEnd() )
				return false;
				
			line = end - 1; //QDocumentLine(*(m_doc->impl()->end() - 1));
			offset = m_doc->line(line).length();
			refreshColumnMemory();
			break;
			
		case QDocumentCursor::StartOfBlock :
			if ( atBlockStart() )
				return false;
				
			m_max = offset = 0;
			break;
			
		case QDocumentCursor::EndOfBlock :
			if ( atBlockEnd() )
				return false;
				
			offset = m_doc->line(line).length();
			refreshColumnMemory();
			break;
			
		case QDocumentCursor::NextBlock :
			
			if ( atEnd() )
				return false;
				
			while ( ((line + 1) < end) && count )
			{
				++line;
				
				if ( !m_doc->line(line).hasFlag(QDocumentLine::Hidden) )
					--count;
					
			}
			
			if ( !count )
			{
				//*line = *it;
				offset = 0;
			} else {
				//*line = QDocumentLine(*(m_doc->impl()->end() - 1));
				offset = m_doc->line(line).length();
			}
			
			break;
			
		case QDocumentCursor::PreviousBlock :
			
			if ( atStart() )
				return false;
				
			offset = 0;
			
			while ( (line > beg) && count )
			{
				--line;
				
				if ( !m_doc->line(line).hasFlag(QDocumentLine::Hidden) )
					--count;
					
			}
			
			if ( !count )
			{
				//*line = *it;
				offset = m_doc->line(line).length();
			} else {
				offset = 0;
				//*line = QDocumentLine(*(m_doc->impl()->begin()));
			}
			
			//*line = *it;
			
			break;
			
		case QDocumentCursor::WordLeft :
		case QDocumentCursor::PreviousWord :
		{
			if ( atStart() )
				return false;
				
			l = m_doc->line(line);
			
			//for ( int loop = 0; loop <= 1; ++loop )
			//{
			//	while ( l.isValid() )
			
			// -- patch --
			/* eats up white space */
			while ( (offset > 0) && !isWord(l.text().at(offset - 1)) )
				--offset;
				
			/* start of line */
			if ( offset == 0 )
			{
				/* first line, first char => nothing to do */
				if( line == beg )
					return true;
					
				do
			// -- patch --
				{
// 					//offset = qMin(offset, l.length() - 1);
// 					bool next = (l.length() && offset >= 0) ? isWord(l.text().at(offset)) : true;
//
// 					if ( loop )
// 						next = !next;
//
// 					if ( !next )
// 						break;
//
// 					if ( offset > 0 )
// 					{
// 						--offset;
// 					} else if ( line != beg ) {
// 						do
// 						{
// 							//*line = *(--it);
// 							--line;
// 							l = m_doc->line(line);
// 							offset = l.length() - 1;
// 						} while ( l.isValid() && (line != beg) && l.hasFlag(QDocumentLine::Hidden) );
// 					} else {
// 						break;
// 					}
// 				}
// 			}
//
// 			while ( l.isValid() )
// 			{
// 				offset = qMin(offset, l.length());
// 				bool next = (offset <= 0) ? false : isWord(l.text().at(offset - 1));
//
// 				if ( !next )
// 					break;
//
// 				--offset;
			
			// -- patch --
					--line;
					l = m_doc->line(line);
					offset = l.length();
				} while ( (line != beg) && l.isValid() && l.hasFlag(QDocumentLine::Hidden) );
				return true;
			// -- patch --
			}
			
			// -- patch --
			/* eats up whole word */
			while ( (offset > 0) && isWord(l.text().at(offset - 1)) )
				--offset;
			// -- patch --
			
			refreshColumnMemory();
			
			break;
		}
		
		case QDocumentCursor::WordRight :
		case QDocumentCursor::NextWord :
		{
			if ( atEnd() )
				return false;
				
			l = m_doc->line(line);
			int lineLength = l.text().length();
			
// 			for ( int loop = 0; loop <= 1; ++loop )
			// -- patch --
			/* end of line */
			if ( offset == lineLength )
 			{
// 				while ( l.isValid() )
				/* last line, last char => nothing to do */
				if ( line == end )
					return true;
			// -- patch --
				do
				{
// 					//offset = qBound(0, offset, l.length() - 1);
// 					bool next = (offset < l.length()) ? isWord(l.text().at(offset)) : true;
//
// 					if ( loop )
// 						next = !next;
//
// 					if ( !next )
// 						break;
//
// 					if ( offset < l.length() )
// 					{
// 						++offset;
// 					} else if ( (line + 1) != end ) {
// 						offset = 0;
// 						do
// 						{
// 							++line;
// 							l = m_doc->line(line);
// 						} while ( l.isValid() && ((line + 1) != end) && (l.hasFlag(QDocumentLine::Hidden) || !l.length()) );
// 					} else {
			// -- patch --
					++line;
					l = m_doc->line(line);
					offset = 0;
				} while ( (line != end) && l.isValid() && l.hasFlag(QDocumentLine::Hidden) );
				
				lineLength = l.text().length();
				/* empty line */
				if ( lineLength == 0 )
					return true;
					
				/* eats up white space */
				while ( !isWord(l.text().at(offset)) )
				{
					++offset;
					/* move to end of line */
					if ( offset == lineLength )
						break;
			// -- patch --
//					}
				}
			// -- patch --
				return true;
			// -- patch --
			}
			
			// -- patch --
			/* next char */
			++offset;
			/* eats up whole word */
			while ( (offset < lineLength) && isWord(l.text().at(offset)) )
				++offset;
				
			/* eats up white space */
			while ( (offset < lineLength) && !isWord(l.text().at(offset)) )
				++offset;
			// -- patch --
			
			refreshColumnMemory();
			
			break;
		}
		
		case QDocumentCursor::StartOfWord :
		{
			int x = wordStart.indexIn(m_doc->line(line).text().left(offset));
			
			if ( x != -1 )
			{
				offset = x;
			} else {
				qDebug("failed to find SOW : %i + %i != %i",
						x, wordStart.matchedLength(), offset);
						
				return false;
			}
			
			refreshColumnMemory();
			
			break;
		}
		
		case QDocumentCursor::EndOfWord :
		{
			int x = wordEnd.indexIn(m_doc->line(line).text(), offset, QRegExp::CaretAtOffset);
			
			if ( x == offset )
			{
				offset += wordEnd.matchedLength();
			} else {
				qDebug("failed to find EOW");
				return false;
			}
			
			refreshColumnMemory();
			
			break;
		}
		
		default:
			qWarning("Unhandled move operation...");
			return false;
	};
	
	return true;
}

void QDocumentCursorHandle::moveTo(const QDocumentCursor &c)
{
	if ( !c.isValid() || !m_doc )
		return;
		
	m_begLine = c.handle()->m_begLine;
	m_begOffset = c.handle()->m_begOffset;
	
	m_endLine = -1;
	m_endOffset = 0;
	
	refreshColumnMemory();
}

void QDocumentCursorHandle::moveTo(int line, int column)
{
	m_begLine = line;
	m_begOffset = column;
	
	m_endLine = -1;
	m_endOffset = 0;
	
	refreshColumnMemory();
}

void QDocumentCursorHandle::insertText(const QString& s, bool keepAnchor)
{
	if ( !m_doc || s.isEmpty() || m_doc->line(m_begLine).isNull() )
		return;
	
	bool sel = hasSelection();
	
	if ( sel )
	{
		beginEditBlock();
		removeSelectedText(keepAnchor);
	}
	
	QDocumentCommand *command = new QDocumentInsertCommand(
										m_begLine,
										m_begOffset,
										s,
										m_doc
									);
									
	command->setTargetCursor(this);
	command->setKeepAnchor(keepAnchor);
	execute(command);
	
	if ( sel )
		endEditBlock();
}

void QDocumentCursorHandle::eraseLine()
{
	if ( !m_doc )
		return;
		
	QDocumentCommand *command = 0;
	
	if ( m_endLine == -1 )
	{
		command = new QDocumentEraseCommand(
										m_begLine,
										0,
										m_begLine + 1,
										0,
										m_doc
									);
	} else {
		command = new QDocumentEraseCommand(
										qMin(m_begLine, m_endLine),
										0,
										qMax(m_begLine, m_endLine) + 1,
										0,
										m_doc
									);
	}
	
	command->setTargetCursor(this);
	execute(command);
}

QChar QDocumentCursorHandle::nextChar() const
{
	if ( !m_doc )
		return QChar();
	
	QDocumentLine l = m_doc->line(m_begLine);
	
	if ( !l.isValid() )
		return QChar();
	
	return m_begOffset < l.length() ? l.text().at(m_begOffset) : QLatin1Char('\n');
}

QChar QDocumentCursorHandle::previousChar() const
{
	if ( !m_doc || (m_begLine <= 0 && m_begOffset <= 0) )
		return QChar();
	
	QDocumentLine l = m_doc->line(m_begLine);
	
	if ( !l.isValid() || m_begOffset > l.length() )
		return QChar();
	
	return m_begOffset ? l.text().at(m_begOffset - 1) : QLatin1Char('\n');
}

void QDocumentCursorHandle::deleteChar()
{
	if ( !m_doc )
		return;
		
	QDocumentLine l = m_doc->line(m_begLine);
	
	if ( l.isNull() || atEnd() )
		return;
		
	QDocumentCommand *command = 0;
	
	if ( !atLineEnd() )
	{
		command = new QDocumentEraseCommand(
											m_begLine,
											m_begOffset,
											m_begLine,
											m_begOffset + 1,
											m_doc
										);
										
	} else {
		// merge two blocks...
		command = new QDocumentEraseCommand(
											m_begLine,
											m_begOffset,
											m_begLine + 1,
											0,
											m_doc
										);
										
	}
	
	command->setTargetCursor(this);
	command->setUndoOffset(-1);
	execute(command);
}

void QDocumentCursorHandle::deletePreviousChar()
{
	if ( !m_doc )
		return;
		
	QDocumentLine l = m_doc->line(m_begLine);
	
	if ( l.isNull() || atStart() )
		return;
		
	QDocumentCommand *command = 0;
	
	if ( !atLineStart() )
	{
		command = new QDocumentEraseCommand(
											m_begLine,
											m_begOffset - 1,
											m_begLine,
											m_begOffset,
											m_doc
										);
										
	} else {
		// merge two blocks...
		QDocumentLine prev = m_doc->line(m_begLine - 1);
		
		command = new QDocumentEraseCommand(
											m_begLine - 1,
											prev.length(),
											m_begLine,
											m_begOffset,
											m_doc
										);
										
	}
	
	command->setTargetCursor(this);
	execute(command);
}

void QDocumentCursorHandle::execute(QDocumentCommand *c)
{
	if ( !m_doc )
		return;
		
	if ( isSilent() && !c->isSilent() )
		c->setSilent(isSilent());
		
	if ( m_blocks.count() )
	{
		c->redo();
		m_blocks.top()->addCommand(c);
		
	} else if ( m_doc ) {
		//qDebug("Cursor handle executing command : 0x%x", this);
		
		m_doc->execute(c);
	}
}

void QDocumentCursorHandle::beginEditBlock()
{
	m_blocks.push(new QDocumentCommandBlock(m_doc));
}

void QDocumentCursorHandle::endEditBlock()
{
	if ( !m_doc || m_blocks.isEmpty() )
		return;
		
	//qDebug("Cursor handle executing command : 0x%x [block]", this);
	
	QDocumentCommandBlock *block = m_blocks.pop();
	
	// special trick to prevent double redo() while getting rid of
	// bugs occuring in when inserting/erasing in overlapping lines
	// inside a command block
	block->setWeakLock(true);
	
	m_doc->execute(block);
}

QDocumentCursor QDocumentCursorHandle::selectionStart() const
{
	if ( !m_doc )
		return QDocumentCursor();
		
	if ( !hasSelection() )
		return QDocumentCursor(clone());
		
	QDocumentCursor pos(m_doc, m_begLine, m_begOffset),
					anc(m_doc, m_endLine, m_endOffset);
					
	return (pos < anc) ? pos : anc;
}

QDocumentCursor QDocumentCursorHandle::selectionEnd() const
{
	if ( !m_doc )
		return QDocumentCursor();
		
	if ( !hasSelection() )
		return QDocumentCursor(clone());
		
	QDocumentCursor pos(m_doc, m_begLine, m_begOffset),
					anc(m_doc, m_endLine, m_endOffset);
					
	return (pos > anc) ? pos : anc;
}

bool QDocumentCursorHandle::eq(const QDocumentCursorHandle *h)
{
	return (m_begLine == h->m_begLine) && (m_begOffset == h->m_begOffset);
	/*
	if ( !hasSelection() )
		return (m_begLine == h->m_begLine) && (m_begOffset == h->m_begOffset);
		
	return
			(m_begLine == h->m_begLine)
		&&
			(m_begOffset == h->m_begOffset)
		&&
			(m_endLine == h->m_endLine)
		&&
			(m_endOffset == h->m_endOffset)
		;
	*/
}

bool QDocumentCursorHandle::lt(const QDocumentCursorHandle *h)
{
	return
				(m_begLine < h->m_begLine)
			||
				((m_begLine == h->m_begLine) && (m_begOffset < h->m_begOffset))
			;
}

bool QDocumentCursorHandle::gt(const QDocumentCursorHandle *h)
{
	return
				(m_begLine > h->m_begLine)
			||
				((m_begLine == h->m_begLine) && (m_begOffset > h->m_begOffset))
			;
}

QString QDocumentCursorHandle::selectedText() const
{
	if ( !m_doc )
		return QString();
		
	QDocumentLine l1 = m_doc->line(m_begLine), l2 = m_doc->line(m_endLine);
	
	if ( l1.isNull() || l2.isNull() )
		return QString();
		
	QString s;
	
	if ( m_begLine == m_endLine )
	{
		int min = qMin(m_begOffset, m_endOffset),
			max = qMax(m_begOffset, m_endOffset);
			
		s = l1.text().mid(min, max - min);
	} else if ( m_begLine < m_endLine ) {
		s = l1.text().mid(m_begOffset);
		
		int it = m_begLine;
		//QDocumentConstIterator it = m_doc->impl()->index(m_begLine.handle());
		
		while ( ++it < m_endLine )
		{
			s += '\n';
			s += m_doc->line(it).text();
		}
		
		s += '\n';
		s += l2.text().left(m_endOffset);
	} else {
		s = l2.text().mid(m_endOffset);
		
		int it = m_endLine;
		//QDocumentConstIterator it = m_doc->impl()->index(m_endLine.handle());
		
		while ( ++it < m_begLine )
		{
			s += '\n';
			s += m_doc->line(it).text();
		}
		
		s += '\n';
		s += l1.text().left(m_begOffset);
	}
	
	return s;
}

void QDocumentCursorHandle::clearSelection()
{
	if ( m_doc && m_doc->line(m_endLine).isValid() )
	{
		//m_begLine = m_endLine;
		//m_begOffset = m_endOffset;
		
		m_endLine = -1;
		m_endOffset = -1;
	}
}

void QDocumentCursorHandle::replaceSelectedText(const QString& text)
{
	int begline, begcol;
	beginBoundary(begline, begcol);
	
	if ( text.isEmpty() )
	{
		removeSelectedText();
	} else {
		insertText(text, true);
	}
	
	//qDebug("[%i, %i] => ( (%i, %i), (%i, %i) )", begline, begcol, m_begLine, m_begOffset, m_endLine, m_endOffset);
}

void QDocumentCursorHandle::select(QDocumentCursor::SelectionType t)
{
	if ( !m_doc || !m_doc->line(m_begLine).isValid() )
		return;
		
	if ( t == QDocumentCursor::LineUnderCursor )
	{
		movePosition(1, QDocumentCursor::StartOfLine, QDocumentCursor::MoveAnchor);
		movePosition(1, QDocumentCursor::EndOfLine, QDocumentCursor::KeepAnchor);
		
	} else if ( t == QDocumentCursor::WordUnderCursor ) {
		
		movePosition(1, QDocumentCursor::StartOfWord, QDocumentCursor::MoveAnchor);
		movePosition(1, QDocumentCursor::EndOfWord, QDocumentCursor::KeepAnchor);
	}
}

void QDocumentCursorHandle::setSelectionBoundary(const QDocumentCursor& c)
{
	if (
			!m_doc
		||
			(m_begLine == -1)
		||
			(
					(c.lineNumber() == m_begLine)
				&&
					(c.columnNumber() == m_begOffset)
			)
		)
		return;
		
	//qDebug("setting new selection boundary... ");
	
	if ( !hasSelection() )
	{
		m_endLine = m_begLine;
		m_endOffset = m_begOffset;
	}
	
	m_begLine = c.lineNumber();
	m_begOffset = c.columnNumber();
}

bool QDocumentCursorHandle::isWithinSelection(const QDocumentCursor& c) const
{
	if ( !hasSelection() ) //|| c.hasSelection() )
		return false;
		
	int minOff, maxOff, min, max;
	
	if ( m_begLine > m_endLine )
	{
		max = m_begLine;
		maxOff = m_begOffset;
		
		min = m_endLine;
		minOff = m_endOffset;
	} else {
		min = m_begLine;
		minOff = m_begOffset;
		
		max = m_endLine;
		maxOff = m_endOffset;
	}
	
	return (m_begLine == m_endLine)
		?
			(
				(c.lineNumber() == m_begLine)
			&&
				(qMin(m_begOffset, m_endOffset) <= c.columnNumber())
			&&
				(qMax(m_begOffset, m_endOffset) >= c.columnNumber())
			)
		:
			(
				(
					(c.lineNumber() > min)
				&&
					(c.lineNumber() < max)
				)
			||
				(
					(c.lineNumber() == min)
				&&
					(minOff <= c.columnNumber())
				)
			||
				(
					(c.lineNumber() == max)
				&&
					(maxOff >= c.columnNumber())
				)
			)
		;
		
}

/*
	beware when modifying these as their current form handle the special
	case of no selection (m_endLine == -1) and a hasty change may break
	that behavior : no selection -> both boundary are the cursor pos = (m_begLine, m_begOffset)
*/
void QDocumentCursorHandle::beginBoundary(int& begline, int& begcol) const
{
	if ( m_begLine < m_endLine )
	{
		begline = m_begLine;
		begcol = m_begOffset;
	} else {
		begline = m_endLine;
		begcol = (m_begLine == m_endLine && m_begOffset < m_endOffset) ? m_begOffset : m_endOffset;
	}
}

void QDocumentCursorHandle::endBoundary(int& endline, int& endcol) const
{
	if ( m_begLine < m_endLine )
	{
		endline = m_endLine;
		endcol = m_endOffset;
	} else {
		endline = m_begLine;
		endcol = (m_begLine == m_endLine && m_begOffset < m_endOffset) ? m_endOffset : m_begOffset;
	}
}

void QDocumentCursorHandle::boundaries(int& begline, int& begcol, int& endline, int& endcol) const
{
	beginBoundary(begline, begcol);
	endBoundary(endline, endcol);
}

void QDocumentCursorHandle::substractBoundaries(int lbeg, int cbeg, int lend, int cend)
{
	int tlmin, tlmax, tcmin, tcmax;
	
	boundaries(tlmin, tcmin, tlmax, tcmax);

	bool begFirst = tlmin == m_begLine && tcmin == m_begOffset;
	
	if ( tlmax < lbeg || tlmin > lend || (tlmax == lbeg && tcmax < cbeg) || (tlmin == lend && tcmin > cend) )
	{
		// no intersection
		return;
	}
	
	int numLines = lend - lbeg;
	bool beyondBeg = (tlmin > lbeg || (tlmin == lbeg && tcmin >= cbeg));
	bool beyondEnd = (tlmax < lend || (tlmax == lend && tcmax <= cend));
	
	if ( beyondBeg && beyondEnd )
	{
		//qDebug("(%i, %i : %i, %i) erased as in (%i, %i : %i, %i)", tlmin, tcmin, tlmax, tcmax, lbeg, cbeg, lend, cend);
		// cursor erased...
		m_begLine = m_endLine = lbeg;
		m_begOffset = m_endOffset = cbeg;
	} else if ( beyondEnd ) {
		//qDebug("beyond end");
		if ( begFirst )
		{
			m_endLine = lbeg;
			m_endOffset = cbeg;
		} else {
			m_begLine = lbeg;
			m_begOffset = cbeg;
		}
	} else if ( beyondBeg ) {
		//qDebug("beyond beg");
		if ( begFirst )
		{
			m_begLine = lend;
			m_begOffset = cend;
			if ( numLines )
			{
				m_begLine -= numLines;
				m_endLine -= numLines;
			} else {
				m_begOffset = cbeg;
			}
			if ( m_begLine == m_endLine )
				m_endOffset -= (cend - cbeg);
		} else {
			m_endLine = lend;
			m_endOffset = cend;
			if ( numLines )
			{
				m_endLine -= numLines;
				m_begLine -= numLines;
			} else {
				m_endOffset = cbeg;
			}
			if ( m_begLine == m_endLine )
				m_begOffset -= (cend - cbeg);
		}
	} else {
		int off = cend - cbeg;
		
		//qDebug("correcting by %i", off);
		
		if ( begFirst )
		{
			m_endLine -= numLines;
			if ( tlmax == lend )
			{
				m_endOffset -= off;
			}
		} else {
			m_begLine -= numLines;
			if ( tlmax == lend )
			{
				m_begOffset -= off;
			}
		}
	}
	
	//qDebug("(%i, %i : %i, %i) corrected to (%i, %i : %i, %i)", tlmin, tcmin, tlmax, tcmax, m_begLine, m_begOffset, m_endLine, m_endOffset);
}

void QDocumentCursorHandle::intersectBoundaries(int& lbeg, int& cbeg, int& lend, int& cend) const
{
	int tlmin, tlmax, tcmin, tcmax, clmin, clmax, ccmin, ccmax;
	
	boundaries(tlmin, tcmin, tlmax, tcmax);
	clmin = lbeg;
	clmax = lend;
	ccmin = cbeg;
	ccmax = cend;
	
	if ( tlmax < clmin || tlmin > clmax || (tlmax == clmin && tcmax < ccmin) || (tlmin == clmax && tcmin > ccmax) )
	{
		lbeg = cbeg = lend = cend = -1;
		return;
	}
	
	if ( tlmin == clmin )
	{
		lbeg = tlmin;
		cbeg = qMax(tcmin, ccmin);
	} else if ( tlmin < clmin ) {
		lbeg = clmin;
		cbeg = ccmin;
	} else {
		lbeg = tlmin;
		cbeg = tcmin;
	}

	if ( tlmax == clmax )
	{
		lend = tlmax;
		cend = qMin(tcmax, ccmax);
	} else if ( tlmax < clmax ) {
		lend = tlmax;
		cend = tcmax;
	} else {
		lend = clmax;
		cend = ccmax;
	}
}

void QDocumentCursorHandle::intersectBoundaries(QDocumentCursorHandle *h, int& lbeg, int& cbeg, int& lend, int& cend) const
{
	int tlmin, tlmax, tcmin, tcmax, clmin, clmax, ccmin, ccmax;
	
	boundaries(tlmin, tcmin, tlmax, tcmax);
	h->boundaries(clmin, ccmin, clmax, ccmax);
	
	if ( tlmax < clmin || tlmin > clmax || (tlmax == clmin && tcmax < ccmin) || (tlmin == clmax && tcmin > ccmax) )
	{
		lbeg = cbeg = lend = cend = -1;
		return;
	}
	
	if ( tlmin == clmin )
	{
		lbeg = tlmin;
		cbeg = qMax(tcmin, ccmin);
	} else if ( tlmin < clmin ) {
		lbeg = clmin;
		cbeg = ccmin;
	} else {
		lbeg = tlmin;
		cbeg = tcmin;
	}
	
	if ( tlmax == clmax )
	{
		lend = tlmax;
		cend = qMin(tcmax, ccmax);
	} else if ( tlmax < clmax ) {
		lend = tlmax;
		cend = tcmax;
	} else {
		lend = clmax;
		cend = ccmax;
	}
}

bool QDocumentCursorHandle::intersect(const QDocumentCursorHandle *c) const
{
	int tlmin, tlmax, tcmin, tcmax, clmin, clmax, ccmin, ccmax;
	
	boundaries(tlmin, tcmin, tlmax, tcmax);
	c->boundaries(clmin, ccmin, clmax, ccmax);
	
	return !( tlmax < clmin || tlmin > clmax || (tlmax == clmin && tcmax < ccmin) || (tlmin == clmax && tcmin > ccmax) );
}

QDocumentCursor QDocumentCursorHandle::intersect(const QDocumentCursor& c) const
{
	if ( !hasSelection() )
	{
		//if ( c.hasSelection() && c.isWithinSelection(QDocumentCursor(this)) )
		//	return QDocumentCursor(clone());
		
	} else if ( !c.hasSelection() ) {
		
		if ( isWithinSelection(c) )
			return c;
		
	} else {
		QDocumentCursorHandle *h = c.handle();

		int lbeg, lend, cbeg, cend;
		intersectBoundaries(h, lbeg, cbeg, lend, cend);
		
		if ( lbeg != -1 )
		{
			QDocumentCursor c(m_doc, lbeg, cbeg);
			
			if ( lend != -1 && (lbeg != lend || cbeg != cend) )
			{
				c.setSelectionBoundary(QDocumentCursor(m_doc, lend, cend));
			}
			
			return c;
		}
	}
	
	return QDocumentCursor();
}

void QDocumentCursorHandle::removeSelectedText(bool keepAnchor)
{
	if ( !m_doc )
		return;
		
	QDocumentLine l1 = m_doc->line(m_begLine), l2 = m_doc->line(m_endLine);
	
	if ( l1.isNull() || l2.isNull() )
		return;
		
	QDocumentCommand *c;
	
	if ( m_begLine < m_endLine )
	{
		c = new QDocumentEraseCommand(
										m_begLine,
										m_begOffset,
										m_endLine,
										m_endOffset,
										m_doc
									);
		
	} else if ( m_begLine > m_endLine ) {
		c = new QDocumentEraseCommand(
										m_endLine,
										m_endOffset,
										m_begLine,
										m_begOffset,
										m_doc
									);
		
	} else {
		c = new QDocumentEraseCommand(
										m_begLine,
										qMin(m_begOffset, m_endOffset),
										m_endLine,
										qMax(m_begOffset, m_endOffset),
										m_doc
									);
		
	}
	
	c->setKeepAnchor(keepAnchor);
	c->setTargetCursor(this);
	execute(c);
}

//////////////////

/////////////////////////
//	QDocumentPrivate
/////////////////////////

template <class T> T* getStaticDefault() { static T _globStatInst; return &_globStatInst; }

QFont* QDocumentPrivate::m_font = 0;// = QApplication::font();
QFontMetrics* QDocumentPrivate::m_fontMetrics = 0;//(m_font);

int QDocumentPrivate::m_defaultTabStop = 4;
QFormatScheme* QDocumentPrivate::m_defaultFormatScheme = getStaticDefault<QFormatScheme>();

QList<QDocumentPrivate*> QDocumentPrivate::m_documents;

bool QDocumentPrivate::m_fixedPitch;
int QDocumentPrivate::m_ascent;// = m_fontMetrics.ascent();
int QDocumentPrivate::m_descent;// = m_fontMetrics.descent();
int QDocumentPrivate::m_leading;// = m_fontMetrics.leading();
int QDocumentPrivate::m_spaceWidth;// = m_fontMetrics.width(' ');
int QDocumentPrivate::m_lineHeight;// = m_fontMetrics.height();
int QDocumentPrivate::m_lineSpacing;// = m_fontMetrics.lineSpacing();

int QDocumentPrivate::m_leftMargin = 5;
QDocument::WhiteSpaceMode QDocumentPrivate::m_showSpaces = QDocument::ShowNone;
QDocument::LineEnding QDocumentPrivate::m_defaultLineEnding = QDocument::Conservative;

int QDocumentPrivate::m_wrapMargin = 15;

QDocumentPrivate::QDocumentPrivate(QDocument *d)
 : 	m_doc(d),
	m_editCursor(0),
	m_lastGroupId(-1),
	m_constrained(false),
	m_width(0),
	m_height(0),
	m_tabStop(m_defaultTabStop),
	m_formatScheme(0),
	m_language(0),
	m_maxMarksPerLine(0),
	m_lineEnding(m_defaultLineEnding)
{
	m_buffer = new QDocumentBuffer(d);
	m_regions = new QDocumentRegionTree(this);
	
	m_documents << this;
	updateFormatCache();
}

QDocumentPrivate::~QDocumentPrivate()
{
	m_marks.clear();
	m_largest.clear();
	
	m_deleting = true;
	delete m_buffer;
	m_deleting = false;
	
	m_commands.clear();
	
	m_documents.removeAll(this);
}

int QDocumentPrivate::findNextMark(int id, int from, int until)
{
	const int numLines = lineCount();
	
	if ( from < 0 )
		from += numLines;
	
	const QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator e = m_marksLookup.constEnd();
	
	int max = until;
	
	if ( max < 0 )
		max += numLines;
	else if ( max < from )
		max = numLines - 1;
		
	for ( int i = from; i <= max; ++i )
	{
		QDocumentLineHandle *h = at(i);
		
		QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator it = m_marksLookup.constFind(h);
		
		if ( it != e )
		{
			foreach ( QDocumentMark *m, *it )
				if ( m->type() == id )
					return i;
		}
	}
	
	if ( until > 0 && until < from )
	{
		for ( int i = 0; i <= until; ++i )
		{
			QDocumentLineHandle *h = at(i);
			
			QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator it = m_marksLookup.constFind(h);
			
			if ( it != e )
			{
				foreach ( QDocumentMark *m, *it )
					if ( m->type() == id )
						return i;
			}	
		}
	}
	
	return -1;
}

int QDocumentPrivate::findPreviousMark(int id, int from, int until)
{
	const int numLines = lineCount();
	
	if ( from < 0 )
		from += numLines;
		
	if ( until < 0 )
	{
		until += numLines;
	} else if ( until >= numLines ) {
		until = numLines - 1;
	}
	
	const QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator e = m_marksLookup.constEnd();
	
	int min = until;
	
	if ( min > from )
		min = 0;
		
	for ( int i = from; i >= min; --i )
	{
		QDocumentLineHandle *h = at(i);
		
		QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator it = m_marksLookup.constFind(h);
		
		if ( it != e )
		{
			foreach ( QDocumentMark *m, *it )
				if ( m->type() == id )
					return i;
		}
	}
	
	if ( until > 0 && until > from )
	{
		const int start = numLines - 1;
		
		for ( int i = start; i >= until; --i )
		{
			QDocumentLineHandle *h = at(i);
			
			QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator it = m_marksLookup.constFind(h);
			
			if ( it != e )
			{
				foreach ( QDocumentMark *m, *it )
					if ( m->type() == id )
						return i;
			}
		}
	}
	
	return -1;
}

void QDocumentPrivate::execute(QDocumentCommand *cmd)
{
	if ( !cmd )
		return;
		
	//qDebug("adding a command...");
	
	//cmd->setTarget(m_doc);
	
	m_lastModified = QDateTime::currentDateTime();
	
	m_commands.push(cmd);
}

bool QDocumentPrivate::updateWrapBoundaries(QDocumentLineHandle *h, int *span)
{
	QHash<QDocumentLineHandle*, QVector<int> >::iterator it = m_boundaries.find(h);
	
	const QHash<QDocumentLineHandle*, QVector<int> >::const_iterator end = m_boundaries.constEnd();
	
	int oldSpan = 0;
	
	if ( it != end )
		oldSpan = it->count();
	
	QVector<int> wb;
	computeWrapBoundaries(h, wb);
	
	const int n = wb.count();
	
	if ( n )
	{
		if ( it != end )
			*it = wb;
		else
			m_boundaries[h] = wb;
	} else {
		if ( it != end )
			m_boundaries.erase(it);
	}
	
	if ( n || oldSpan )
	{
		if ( h->hasFlag(QDocumentLine::UseLayout) )
			h->setFlag(QDocumentLine::TextChanged, true);
		else
			h->setFlag(QDocumentLine::FormattingChanged, true);
	}
	
	if ( span )
		*span = n;
	
	return oldSpan != n;
}

void QDocumentPrivate::computeWrapBoundaries(QDocumentLineHandle *h, QVector<int>& m_boundaries)
{
	if ( !m_constrained )
		return;
	
	//qDebug("computing wrap boundaries for 0x%x", h);
	
	if ( h->hasFlag(QDocumentLine::UseLayout) )
	{
		// FIXME :
	} else {
		const int sw = m_spaceWidth;
		const int len  = h->m_text.length();
		
		if ( m_fixedPitch )
		{
			if ( !h->hasFlag(QDocumentLine::HasTabs) || m_tabStop == 1 )
			{
				if ( m_leftMargin + len * sw < m_width )
				{
					return;
				}
			} else {
				
				// TODO : be smarter...
			}
		} else {
			
			// TODO : implement fast bypasses
		}
		
		int idx = 0, minx = 0, lastBreak = 0, lastWidth = 0,
			x = m_leftMargin, column = 0, cwidth;
		
		const QChar * const d = h->m_text.constData();
		const quint16 * unicode = h->m_text.utf16();
		
		while ( idx < len )
		{
			const quint16 uni = *unicode; //unicode[idx]; //h->m_text.at(idx).unicode();
			
			//c = m_text.at(idx);
			if ( uni > ' ' )
				break;
			
			if ( uni == '\t' )
			{
				const int taboffset = m_tabStop - (column % m_tabStop);
				
				column += taboffset;
				
				cwidth = sw * taboffset;
			} else {
				++column;
				cwidth = sw;
			}
			
			x += cwidth;
			++unicode;
			++idx;
		}
		
		if ( (x + sw) >= m_width )
		{
			//qWarning("Please stop shrinking so aggressively.\nNo attempt will be made to show something decent");
			// shrinking too aggresively (or too much spaces...) ungraceful fallback
			
			idx = 0;
			x = m_leftMargin;
		}
		
		minx = x;
		lastBreak = idx;
		
		const int anchor = idx;
		//m_indent = minx - QDocumentPrivate::m_leftMargin;
		
		while ( idx < len )
		{
			//c = h->m_text.at(idx);
			
			const quint16 uni = *unicode; //unicode[idx]; //c.unicode();
			
			if ( uni <= ' ' ) //!isWord(c) || !isWord(m_text.at(idx)) )
			{
				lastWidth = x + sw;
				lastBreak = idx + 1;
			}
			
			if ( uni == '\t' )
			{
				const int taboffset = m_tabStop - (column % m_tabStop);
				
				column += taboffset;
				cwidth = sw * taboffset;
			} else {
				++column;
				
				if ( !QDocumentPrivate::m_fixedPitch )
				{
					// FIXME : get format info from cache?
					//QFontMetrics fm = *m_fontMetrics; //(fonts.at(fmt));
					cwidth = m_fontMetrics->width(d[idx]);
				} else {
					cwidth = sw;
				}
			}
			
			if ( x + cwidth > m_width )
			{
				if ( lastBreak == (m_boundaries.count() ? m_boundaries.last() : anchor) )
				{
					// perfect cut or fallback to aggressive cut
					//m_frontiers << qMakePair(idx, rx);
					if ( uni == ' ' )
					{
						m_boundaries << idx;
						lastBreak = idx;
						lastWidth = x + cwidth;
					} else {
						m_boundaries << idx - 1;
						lastBreak = idx - 1;
						lastWidth = x;
					}
					x = minx;
				} else if ( lastBreak && lastBreak < idx ) {
					// word cut at a non-ideal position
					//m_frontiers << qMakePair(lastBreak, lastX);
					m_boundaries << lastBreak;
					x = minx + (x - lastWidth);
				} else {
					m_boundaries << idx;
					lastBreak = idx;
					lastWidth = x;
					x = minx;
				}
				
				//qDebug("boundary at : %i", idx);
			}
			
			x += cwidth;
			++unicode;
			++idx;
		}
		
		// update format ranges
		// must be done outside to make sure the boundary data is stored
		//compose(h);
	}
}

int QDocumentPrivate::wrappedLineForCursor(QDocumentLineHandle *h, int cpos)
{
	if ( !m_constrained )
		return 0;
	
	QVector<int> m_boundaries;
	
	m_boundaries = wrapBoundaries(h);
	
	int wrap = m_boundaries.count();
	
	for ( int i = 0; i < wrap; ++i )
	{
		if ( m_boundaries.at(i) > cpos )
		{
			wrap = i;
			break;
		}
	}
	
	return wrap;
}

int QDocumentPrivate::documentOffsetToCursor(QDocumentLineHandle *h, int x, int y)
{
	int wrap = y / m_lineSpacing;
	
	QVector<int> m_boundaries;
	
	if ( m_constrained )
		m_boundaries = wrapBoundaries(h);
	
	if ( wrap > m_boundaries.count() )
	{
		// return an invalid value instead?
		//qDebug("a bit too far : (%i, %i)", x, y);
		//wrap = m_frontiers.count();
		
		return h->m_text.length();
	}
	
	if ( m_boundaries.count() )
	{
		//qDebug("(%i, %i) : %i", x, y, wrap);
		x = qMin(x, m_width);
	}
	
	int cpos = 0;
	int max = h->m_text.length();
	
	if ( wrap < m_boundaries.count() )
		max = m_boundaries.at(wrap) - 1;
	
	if ( wrap > 0 )
		cpos = m_boundaries.at(wrap - 1);
	
	x -= m_leftMargin;
	
	const int ts = m_tabStop;
	int idx = cpos, column = 0;
	
	if ( h->hasFlag(QDocumentLine::UseLayout) )
	{
		cpos = h->m_render.layout->lineAt(wrap).xToCursor(x);
	} else if ( QDocumentPrivate::m_fixedPitch ) {
		
		// FIXME : compute indent
		if ( wrap )
			x -= m_leftMargin; //m_indent;
		
		while ( (idx < max) && (x > 0) )
		{
			bool tab = h->m_text.at(idx).unicode() == '\t';
			int cwidth = m_spaceWidth;
			
			if ( tab )
			{
				int coff = ts - (column % ts);
				column += coff;
				cwidth *= coff;
			} else {
				++column;
			}
			
			int thresold = (2 * cwidth) / 3;
			
			if ( x <= thresold )
				break;
				
			x -= cwidth;
			++idx;
		}
		
		cpos = idx;
	} else {
		// FIXME unconstify the method?
		//if ( hasFlag(QDocumentLine::FormattingChanged) )
		//	compose();
		
		// FIXME : compute indent
		if ( wrap )
			x -= m_leftMargin; //m_indent;
		
		while ( (idx < max) && (x > 0) )
		{
			// FIXME : get formatting from render range cache
			int fid = 0; //idx < m_cache.count() ? m_cache[idx] : 0;
			
			if ( fid < 0 )
				fid = 0;
			
			QFontMetrics fm = (fid < m_fonts.count()) ? QFontMetrics(m_fonts.at(fid)) : m_doc->fontMetrics();
			
			QChar c = h->m_text.at(idx);
			bool tab = c.unicode() == '\t';
			int cwidth = 0;
			
			if ( tab )
			{
				int coff = ts - (column % ts);
				column += coff;
				cwidth = fm.width(' ');
				cwidth *= coff;
			} else {
				++column;
				cwidth = fm.width(c);
			}
			
			int thresold = (2 * cwidth) / 3;
			
			if ( x <= thresold )
				break;
				
			x -= cwidth;
			++idx;
		}
		
		cpos = idx;
	}
	
	return cpos;
}

void QDocumentPrivate::cursorToDocumentOffset(QDocumentLineHandle *h, int cpos, int& x, int& y)
{
	QVector<int> m_boundaries;
	
	if ( m_constrained )
		m_boundaries = wrapBoundaries(h);
	
	if ( cpos > h->m_text.length() )
		cpos = h->m_text.length();
	else if ( cpos < 0 )
		cpos = 0;
		
	int idx = 0;
	int wrap = wrappedLineForCursor(h, cpos);
	
	x = m_leftMargin;
	y = wrap * m_lineSpacing;
	
	if ( wrap )
	{
		idx = m_boundaries.at(wrap - 1);
	}
	
	int column = 0;
	const int ts = m_tabStop;
	
	if ( h->hasFlag(QDocumentLine::UseLayout) )
	{
		x += h->m_render.layout->lineAt(wrap).cursorToX(cpos);
	} else {
		if ( !QDocumentPrivate::m_fixedPitch )
		{
			if ( h->hasFlag(QDocumentLine::FormattingChanged) )
				compose(h);
		}
		
		// FIXME : compute indent
		if ( wrap )
			x -= m_leftMargin; //m_indent;
		
		while ( idx < cpos )
		{
			bool tab = h->m_text.at(idx).unicode() == '\t';
			int cwidth = QDocumentPrivate::m_spaceWidth;
			
			if ( tab )
			{
				int coff = ts - (column % ts);
				column += coff;
				cwidth *= coff;
			} else {
				++column;
				
				if ( !m_fixedPitch )
				{
					int fid = 0; // TODO : figure out format from render range cache
					QFontMetrics fm = (fid < m_fonts.count()) ? QFontMetrics(m_fonts.at(fid)) : *m_fontMetrics;
					cwidth = fm.width(h->m_text.at(idx));
				}
			}
			
			x += cwidth;
			++idx;
		}
	}
}

QPoint QDocumentPrivate::cursorToDocumentOffset(QDocumentLineHandle *h, int cpos)
{
	QPoint p;
	cursorToDocumentOffset(h, cpos, p.rx(), p.ry());
	return p;
}

/*
	Update the render range cache
	
	QVector<int> :
		n = size of range
		n + 1 = format (packed) :
				* sign bit (0x80000000) indicates bunch of spaces (with tabs in it...)
				* next bit (0x40000000) indicates composite format
				* data (0x0fffffff)
					** if composite format : number of format to merge
					** format id
		n + ? = formats to merge upon drawing, if any
	
	selections are merged on-the-fly upon drawing
*/

void QDocumentPrivate::compose(QDocumentLineHandle *h)
{
	if ( h->hasFlag(QDocumentLine::UseLayout) || h->m_text.isEmpty() )
		return;
	
	if ( !h->m_render.data )
	{
		h->m_render.data = new QVector<int>();
	}	else {
		h->m_render.data->clear();
	}
	
	QVector<int> *d = h->m_render.data;
	d->reserve(10);
	
	QVector<int> m_boundaries;
	
	if ( m_constrained )
		m_boundaries = wrapBoundaries(h);
	
	const int cxtCount = h->m_contexts.count();
	
	QStack<int> cxtFmts;
	QList<int> enter, leave;
	QStack<QLanguageDefinition*> ld;
	
	cxtFmts.push(0);
	ld.push(m_language);
	
	//qDebug() << h->m_contexts;
	
	for ( int i = 0; i < cxtCount; i += 3 )
	{
		int idx = h->m_contexts.at(i);
		int len = h->m_contexts.at(i + 1);
		int cid = h->m_contexts.at(i + 2);
		
		if ( idx == -1 )
		{
			if ( cid & QLanguageDefinition::Language )
			{
				ld.push(QLanguageDefinition::instance(cid));
			} else {
				cxtFmts.push(ld.top()->contextFormat(cid));
			}
		} else {
			enter << idx << cid;
		}
		
		if ( len != -1 )
		{
			int j = 0;
			int sidx = qMax(0, idx) + len;
			
			for ( ; (j + 1) < leave.count(); j += 2 )
			{
				if ( leave.at(j) > sidx )
					break;
			}
			
			leave.insert(j, cid);
			leave.insert(j, sidx);
		}
	}
	
	int tokenIdx = 1, overlayIdx = h->m_tokens.count() ? h->m_tokens.first() : 0;
	
	int i = 0, wrap = 0, max = h->m_text.count();
	const QChar *td = h->m_text.constData();
	
	while ( i < max )
	{
		int spaceBoundary = i;
		bool isInSpaceChunk = isspace(td[i].unicode());
		int wrapBoundary = wrap < m_boundaries.count() ? m_boundaries.at(wrap) : max;
		
		int cxtStartBoundary = enter.count() ? enter.first() : max;
		int cxtStopBoundary = leave.count() ? leave.first() : max;
		
		if ( isInSpaceChunk )
		{
			do
			{
				++spaceBoundary;
			} while ( spaceBoundary < max && td[spaceBoundary].unicode() <= ' ' ); //.isSpace() );
		} else {
			do
			{
				++spaceBoundary;
			} while ( spaceBoundary < max && td[spaceBoundary].unicode() > ' ' ); //isSpace() ); //.unicode() != '\t') )
		}
		
		int bound = qMin(spaceBoundary, wrapBoundary);
		
		int cxtBound = qMin(cxtStartBoundary, cxtStopBoundary);
		
		int tokenStart = (tokenIdx + 2) < overlayIdx ? h->m_tokens.at(tokenIdx) : max;
		
		//qDebug("%i, %i, %i, %i, %i", spaceBoundary, wrapBoundary, cxtStartBoundary, cxtStopBoundary, tokenStart);
		
		bool cxtStart = i < max && cxtStartBoundary == i;
		bool cxtStop = i < max && cxtStopBoundary == i;
		
		if ( cxtStop )
		{
			// change default format
			if ( leave.at(1) & QLanguageDefinition::Language )
			{
				ld.pop();
			} else {
				cxtFmts.pop();
			}
			
			leave.removeFirst();
			leave.removeFirst();
		}
		
		if ( cxtStart )
		{
			// change default format
			int cid = enter.at(1);
			
			if ( cid & QLanguageDefinition::Language )
			{
				ld.push(QLanguageDefinition::instance(cid));
				//format = ;
			} else {
				cxtFmts.push(ld.top()->contextFormat(cid));
			}
			
			//qDebug("%i -> %i", cxtFmts.top(), format);
			
			enter.removeFirst();
			enter.removeFirst();
		}
		
		if ( tokenStart == i )
		{
			int tokenLength = h->m_tokens.at(++tokenIdx);
			int tokenFormat = h->m_tokens.at(++tokenIdx);
			++tokenIdx;
			
			if ( ld.count() )
				tokenFormat = ld.top()->tokenFormat(tokenFormat);
			
			int end = tokenStart + tokenLength;
			
			//qDebug("token:(%i, %i, %i)", tokenStart, tokenLength, tokenFormat);
			
			if ( end <= bound )
			{
				//qDebug("full token added :D");
				d->append(tokenLength);
				d->append(tokenFormat);
			} else {
				//qDebug("token:(%i, %i, %i)", tokenStart, tokenLength, tokenFormat);
				
				int chunk = bound - tokenStart;
				d->append(chunk);
				d->append(tokenFormat);
				tokenLength -= chunk;
				
				//qDebug("chunk:%i", chunk);
				
				while ( tokenLength > 0 && bound < end )
				{
					if ( bound == wrapBoundary )
						++wrap;
					
					isInSpaceChunk = td[bound].unicode() <= ' '; //.isSpace(); //.unicode() == '\t';
					wrapBoundary = wrap < m_boundaries.count() ? m_boundaries.at(wrap) : max;
					
					if ( isInSpaceChunk )
					{
						do
						{
							++spaceBoundary;
						} while ( spaceBoundary < max && td[spaceBoundary].unicode() <= ' ' ); //.isSpace() );
					} else {
						do
						{
							++spaceBoundary;
						} while ( spaceBoundary < max && td[spaceBoundary].unicode() > ' ' ); //.isSpace() );
					}
					
					i = bound;
					bound = qMin(end, qMin(spaceBoundary, wrapBoundary));
					
					chunk = bound - i;
					
					if ( chunk )
					{
						d->append(chunk);
						d->append(tokenFormat | (isInSpaceChunk ? 0x80000000 : 0));
						tokenLength -= chunk;
					} else {
						break;
					}
					//qDebug("chunk:%i", chunk);
				}
			}
			
			i = end;
		} else {
			bound = qMin(bound, cxtBound);
			bound = qMin(bound, tokenStart);
			
			if ( bound > i )
			{
				//qDebug("range:(%i, %i, %i)", i, bound - i, cxtFmts.top());
				
				d->append(bound - i);
				d->append(cxtFmts.top() | (isInSpaceChunk ? 0x80000000 : 0));
				i = bound;
			}
			
			if ( bound >= wrapBoundary )
				++wrap;
		}
	}
	
	// note : do that upon render instead?
	// TODO : change overlay storage to allow a smarter implementation of overlays merging
	
	while ( overlayIdx + 2 < h->m_tokens.count() )
	{
		int idx = h->m_tokens.at(overlayIdx);
		int len = h->m_tokens.at(++overlayIdx);
		int fmt = h->m_tokens.at(++overlayIdx);
		
		++overlayIdx;
		
		int pos = 0;
		int range = 0;
		
		while ( range + 1 < d->count() )
		{
			int rlen = d->at(range);
			
			if ( idx < pos )
				break;
			
			int rfmt = d->at(++range);
			
			if ( idx < pos + rlen )
			{
				// merge...
				int off = idx - pos;
				int common = rlen - off;
				
				if ( !off && len == rlen )
				{
					// TODO : merge...
					(*d)[range] = fmt;
					break;
				}
				
				if ( off )
				{
					// split
					(*d)[range - 1] = off;
					
					// TODO : merge...
					d->insert(++range, common);
					d->insert(++range, fmt);
				} else {
					(*d)[range] = fmt;
				}
				
				int diff = len - common;
				
				if ( diff < 0 )
				{
					//
					(*d)[range - 1] += diff;
					d->insert(++range, -diff);
					d->insert(++range, rfmt);
				} else if ( diff > 0 ) {
					//
					idx += common;
					len = diff;
				} else {
					break;
				}
			}
			
			pos += rlen;
			++range;
		}
	}
	
	h->setFlag(QDocumentLine::FormattingChanged, false);
}

void QDocumentPrivate::layout(QDocumentLineHandle *h)
{
	bool needLayout = false; //!(m_text.isRightToLeft() && m_text.isSimpleText());
	
	/*
	static const QList<QChar::Direction> m_layoutRequirements = QList<QChar::Direction>()
		<< QChar::DirR
		<< QChar::DirAL
		<< QChar::DirRLE
		<< QChar::DirRLO
		<< QChar::DirPDF
		<< QChar::DirAN;
	*/
	
	h->m_state &= ~QDocumentLine::HasTabs;
	
	int len = h->m_text.length();
	const QChar *d = h->m_text.constData();
	
	while ( len > 0 )
	{
		quint16 uni = d->unicode();
		
		// try to be smart : 
		//  * only stoopid out-of-ASCII range char may need checking
		//  * calling direction() can be expensive (in large amounts)
		if ( uni & 0xff80 )
		{
			//needLayout = m_layoutRequirements.contains(d[i].direction());
			
			QChar::Direction dir = d->direction();
			
			if (
					(dir == QChar::DirR) 
				||
					(dir == QChar::DirRLE)
				||
					(dir == QChar::DirRLO) 
				||
					(dir == QChar::DirAL)
				||
					(dir == QChar::DirAN)
				||
					(dir == QChar::DirPDF)
				)
			{
				needLayout = true;
				break;
			}
		} else if ( uni == '\t' ) {
			h->m_state |= QDocumentLine::HasTabs;
		}
		
		--len;
		++d;
	}
	
	if ( needLayout )
	{
		//qDebug("layout needed at line %i", this->line());
		
		if ( !h->hasFlag(QDocumentLine::UseLayout) )
		{
			h->m_render.layout = new QTextLayout(h->m_text, QDocument::font());
			h->setFlag(QDocumentLine::UseLayout);
		} else {
			h->m_render.layout->setText(h->m_text);
			//m_layout->setFont(config()->font());
		}
		
		h->m_render.layout->setCacheEnabled(false); //true); //cacheLayout);
		
		// Initial setup of the QTextLayout.
		
		// Tab width
		QTextOption opt;
		opt.setFlags(QTextOption::IncludeTrailingSpaces);
		opt.setTabStop(m_doc->tabStop() * QDocumentPrivate::m_spaceWidth);
		
		opt.setWrapMode(QTextOption::WrapAtWordBoundaryOrAnywhere);
		
		// Find the first strong character in the string.
		// If it is an RTL character, set the base layout direction of the string to RTL.
		//
		// See http://www.unicode.org/reports/tr9/#The_Paragraph_Level (Sections P2 & P3).
		// Qt's text renderer ("scribe") version 4.2 assumes a "higher-level protocol"
		// (such as KatePart) will specify the paragraph level, so it does not apply P2 & P3
		// by itself. If this ever change in Qt, the next code block could be removed.
		if ( h->m_text.isRightToLeft() )
			opt.setTextDirection(Qt::RightToLeft);
			
		h->m_render.layout->setTextOption(opt);
		
		// Syntax highlighting, inbuilt and arbitrary
		//m_layout->setAdditionalFormats(m_ranges);
		
		// Begin layouting
		h->m_render.layout->beginLayout();
		
		int i = 0, rx = 0, height = 0;
		
		forever
		{
			QTextLine line = h->m_render.layout->createLine();
			
			if ( !line.isValid() )
			{
				//if ( m_layout->lineCount() > 1 )
				//	qWarning("Troubles expected on line %i", this->line());
				
				break;
			}
			
			if ( m_constrained )
				line.setLineWidth(m_width - QDocumentPrivate::m_leftMargin);
			else
				line.setNumColumns(h->m_text.length());
				
			rx += qRound(line.naturalTextWidth());
			
			if ( m_constrained && h->m_render.layout->textOption().textDirection() == Qt::RightToLeft )
			{
				line.setPosition(QPoint(qRound(qreal(m_width - 2 * QDocumentPrivate::m_leftMargin) - line.naturalTextWidth()), height));
			} else {
				line.setPosition(QPoint(0, height));
			}
			
			++i;
			
			height += m_lineSpacing;
		}
		
		h->m_render.layout->endLayout();
	} else {
		if ( h->hasFlag(QDocumentLine::UseLayout) )
			delete h->m_render.layout;
		
		h->setFlag(QDocumentLine::UseLayout, false);
		h->setFlag(QDocumentLine::FormattingChanged, true);
		h->m_render.data = 0;
	}
	
	h->setFlag(QDocumentLine::TextChanged, false);
}

void QDocumentPrivate::draw(	QDocumentLineHandle *h,
								QPainter *p,
								int xOffset,
								int vWidth,
								const QVector<int>& sel,
								const QVector<int>& cursor,
								const QPalette& pal,
								bool fullSel)
{
	if ( h->hasFlag(QDocumentLine::TextChanged) )
		layout(h);
	
	//qDebug("draw : 0x%x", h);
	
	if ( h->hasFlag(QDocumentLine::UseLayout) )
	{
		if ( h->hasFlag(QDocumentLine::FormattingChanged) )
		{
			h->m_render.layout->setAdditionalFormats(h->decorations());
			h->setFlag(QDocumentLine::FormattingChanged, false);
		}
		
		const int lh = m_fontMetrics->height();
		
		QVector<QTextLayout::FormatRange> selections;
		
		QTextCharFormat fmt;
		fmt.setBackground(pal.highlight());
		fmt.setForeground(pal.highlightedText());
		
		QTextLayout::FormatRange range;
		if ( fullSel )
		{
			range.start = 0;
			range.format = fmt;
			range.length = h->m_text.length();
			selections << range;
		} else {
			for ( int i = 0; i < sel.count(); ++i )
			{
				range.start = sel[i];
				range.format = fmt;
				
				if ( (i + 1) < sel.count() )
				{
					// regular selection subset
					range.length = sel[++i] - range.start;
					
				} else if ( h->m_render.layout->lineCount() ) {
					// span to end of line, not only text
					range.length = h->m_text.length() - range.start;
					qreal lineWidth = h->m_render.layout->lineAt(h->m_render.layout->lineCount() - 1).naturalTextWidth();
					const int endX = m_leftMargin + qRound(lineWidth) - xOffset;
					
					QRect area(endX, lh * i, vWidth - endX, lh);
					
					p->fillRect(area, fmt.background());
				}
				
				selections << range;
			}
		}
		
		const QPoint off(QDocumentPrivate::m_leftMargin, 0);
		
		//m_layout->setAdditionalFormats(decorations());
		
		//qDebug("%i, %i", m_layout->additionalFormats().count(), selections.count());
		
		h->m_render.layout->draw(p, off, selections);
		
		for ( int i = 0; i < cursor.count(); ++i )
			h->m_render.layout->drawCursor(p, off, cursor[i]);
			
		//m_layout->clearAdditionalFormats();
	} else if ( h->m_text.isEmpty() ) {
		// enforce selection drawing on empty lines
		if ( sel.count() == 1 )
			p->fillRect(
						qMax(xOffset, m_leftMargin),
						0,
						vWidth,
						m_lineSpacing,
						pal.highlight()
						);
						
		// enforce cursor drawing on empty lines
		if ( cursor.count() && (xOffset < m_leftMargin) )
			p->drawLine(
						m_leftMargin,
						0,
						m_leftMargin,
						m_lineSpacing
						);
		
	} else {
		
		QVector<int> m_boundaries;
		
		if ( m_constrained )
			m_boundaries = wrapBoundaries(h);
		
		if ( h->hasFlag(QDocumentLine::FormattingChanged) )
		{
			compose(h);
		}
		
		const int ts = m_tabStop;
		const int maxWidth = xOffset + vWidth;
		const bool unbounded = sel.count() & 1;
		const QColor ht = pal.highlightedText().color();
		
		int last = h->previousNonSpaceChar(h->m_text.length() - 1);
		
		const bool showTabs = QDocument::showSpaces() & QDocument::ShowTabs,
				showLeading = QDocument::showSpaces() & QDocument::ShowLeading,
				showTrailing = QDocument::showSpaces() & QDocument::ShowTrailing;
				
		//const int fns = nextNonSpaceChar(0);
		int indent = m_leftMargin; //qMax(m_indent, QDocumentPrivate::m_leftMargin);
		
		int cidx = 0;
		int selidx = 0;
		int rngIdx = 0;
		int column = 0;
		int textPos = 0;
		bool continuingWave = false, brokenWave = false;
		int dir = 0; // 0 = down; 1 = up
		int wrap = 0, xpos = m_leftMargin, ypos = 0;
		
		bool inSel = false;
		const int maxRange = h->m_render.data->count();
		
		const int firstFmt = maxRange > 1 ? h->m_render.data->at(1) : 0;
		
		int fmt = (firstFmt & 0xffffff) + 1;
		bool leading = firstFmt & 0x80000000, pastLead = false;
		
		while ( rngIdx + 1 < maxRange )
		{
			int rngLen = h->m_render.data->at(rngIdx);
			int rngFmt = h->m_render.data->at(++rngIdx);
			++rngIdx;
			
			if ( wrap < m_boundaries.count() && textPos >= m_boundaries.at(wrap) )
			{
				continuingWave = false;
				
				if ( inSel && !fullSel )
				{
					// finish selection
					p->fillRect(
									xpos, ypos,
									maxWidth - xpos, m_lineSpacing,
									pal.highlight()
								);
								
				}
				
				if ( pastLead && (rngFmt & 0x80000000) )
				{
					indent = m_leftMargin;
				}
				
				++wrap;
				column = 0;
				ypos += m_lineSpacing;
				xpos = indent;
				
				if ( inSel && !fullSel )
				{
					// finish selection
					p->fillRect(
									m_leftMargin, ypos,
									xpos, m_lineSpacing,
									pal.highlight()
								);
								
				}
			}
			
			pastLead = false;
			
			// TODO : format merging
			
			// TODO : clip more accurately (i.e inside ranges)
			if ( xpos > maxWidth )
				break;
			
			if ( (rngFmt & 0xffffff) != fmt )
			{
				tunePainter(p, rngFmt & 0xffffff);
				
				fmt = rngFmt & 0xffffff;
			}
			
			int rwidth = 0;
			int tcol = column;
			const QString rng = h->m_text.mid(textPos, rngLen);
			int prevTextPos = textPos;
			textPos += rngLen;
			
			if ( rngFmt & 0x80000000 )
			{
				foreach ( QChar c, rng )
				{
					if ( c.unicode() == '\t' )
					{
						int toff = ts - (tcol % ts);
 						rwidth += toff * m_spaceWidth;
						tcol += toff;
					} else {
						rwidth += m_spaceWidth;
						++tcol;
					}
				}
			} else {
				column += rngLen;
				
				if ( m_fixedPitch )
					rwidth = m_spaceWidth * rngLen;
				else
					rwidth = p->fontMetrics().width(rng);
				
				leading = false;
			}
			
			if ( (xpos + rwidth) <= xOffset )
			{
				//qDebug("skipped");
				xpos += rwidth;
				
				if ( rngFmt & 0x80000000 )
					column = tcol;
				
				continue;
			}
			
			QVector<int> selOverlap;
			
			QFormat format;
			int xspos = xpos;
			const QPen oldpen = p->pen();
			const int baseline = ypos + m_ascent;
			
			if ( m_formatScheme )
				format = m_formatScheme->format(fmt);
			else if ( m_defaultFormatScheme )
				format = m_defaultFormatScheme->format(fmt);
			
			if ( fullSel )
			{
				p->setPen(ht);
				
				p->fillRect(xpos, ypos,
							rwidth, m_lineSpacing,
							pal.highlight());
				
			} else {
				//p->setPen((rngIdx >> 1) & 1 ? Qt::red : Qt::darkGreen);
				
				if ( format.foreground.isValid() )
				{
					p->setPen(format.foreground);
				} else {
					p->setBrush(pal.text());
				}
				
				
				// TODO : only paint the background for unselected parts?
				if ( format.background.isValid() )
				{
					p->fillRect(xpos, ypos,
								rwidth, m_lineSpacing,
								format.background);
				}
				
				while ( selidx < sel.count() )
				{
					int selBeg = sel.at(selidx);
					int selEnd = selidx + 1 < sel.count() ? sel.at(selidx + 1) : -1; 
					
					if ( selBeg > textPos )
						break;
					
					if ( selEnd != -1 && selEnd < prevTextPos )
					{
						selidx += 2;
						continue;
					}
					
					int start = qMax(prevTextPos, selBeg);
					int stop = selEnd != -1 ? qMin(textPos, selEnd) : textPos;
					int len = stop - start;
					
					int selx, selw;
					
					selOverlap << start;
					selOverlap << len;
					
					if ( m_fixedPitch )
					{
						selx = xpos + (start - prevTextPos) * m_spaceWidth;
						selw = len * m_spaceWidth;
					} else {
						selx = xpos + p->fontMetrics().width(h->m_text.mid(prevTextPos, start - prevTextPos));
						selw = p->fontMetrics().width(h->m_text.mid(start, len));
					}
					
					if ( !(rngFmt & 0x80000000) )
					{
						p->fillRect(selx, ypos,
									selw, m_lineSpacing,
									pal.highlight());
						
					}
					
					if ( stop == selEnd )
					{
						selidx += 2;
					} else {
						break;
					}
				}
			}
			
			if ( rngFmt & 0x80000000 )
			{
				// spaces
				int max = textPos;
				
				if ( rngIdx + 1 >= h->m_render.data->count() )
					++max;
				
				int selOverlapIdx = 0;
				
				for ( int i = prevTextPos; i < max; ++i )
				{
					while ( cidx < cursor.count() )
					{
						if ( cursor.at(cidx) == i )
						{
							p->drawLine(xpos, ypos, xpos, ypos + m_lineSpacing);
							
							++cidx;
						} else {
							break;
						}
					}
					
					if ( i == textPos )
						break;
					
					bool isTab = h->m_text.at(i).unicode() == '\t';
					
					bool isInSel = false;
					if ( selOverlapIdx < selOverlap.count() )
					{
						// TODO : could do with some optimization
						int start = selOverlap.at(selOverlapIdx);
						
						if ( start <= i )
						{
							int len = selOverlap.at(selOverlapIdx + 1);
							
							if ( start + len <= i )
							{
								selOverlapIdx += 2;
							} else {
								isInSel = true;
							}
						}
					}
					
					if ( isInSel )
					{
						p->save();
						p->setPen(ht);
					}
					
					if ( isTab )
					{
						int toff = ts - (column % ts);
						column += toff;
						int xoff = toff * QDocumentPrivate::m_spaceWidth;
						
						if ( isInSel )
							p->fillRect(xpos, ypos,
										xoff, m_lineSpacing,
										pal.highlight());
						
						if ( showTabs )
						{
							p->translate(xpos - m_spaceSignOffset, ypos + m_ascent);
							p->drawPoints(m_spaceSign, sizeof(m_spaceSign) / sizeof(QPoint));
							p->translate(m_spaceSignOffset - xpos,-ypos - m_ascent);
						}
						
						xpos += xoff;
					} else {
						++column;
						
						if ( isInSel )
							p->fillRect(xpos, ypos,
										m_spaceWidth, m_lineSpacing,
										pal.highlight());
						
						if (
								(
									leading
								&&
									showLeading
								)
							||
								(
									(prevTextPos >= last)
								&&
									showTrailing
								)
							)
						{
							p->translate(xpos - m_spaceSignOffset, ypos + m_ascent);
							p->drawPoints(m_spaceSign, sizeof(m_spaceSign) / sizeof(QPoint));
							p->translate(m_spaceSignOffset - xpos,-ypos - m_ascent);
						}
						
						xpos += m_spaceWidth;
					}
					
					if ( isInSel )
					{
						p->restore();
					}
				}
				
				if ( leading )
				{
					//indent = xpos;
					leading = false;
					pastLead = true;
				}
				
			} else {
				if ( selOverlap.isEmpty() )
				{
					p->drawText(xpos, baseline, rng);
				} else {
					int last = prevTextPos;
					
					int fxpos = xpos;
					const QPen bak = p->pen();
					
					for ( int oi = 0; oi < selOverlap.count(); ++oi )
					{
						QString tmp;
						int next = selOverlap.at(oi);
						
						if ( last != next )
						{
							// draw unselected
							tmp = h->m_text.mid(last, next - last);
							p->drawText(fxpos, baseline, tmp);
							fxpos += p->fontMetrics().width(tmp);
						}
						
						int len = selOverlap.at(++oi);
						
						// draw selected
						p->setPen(ht);
						tmp = h->m_text.mid(next, len);
						p->drawText(fxpos, baseline, tmp);
						fxpos += p->fontMetrics().width(tmp);
						p->setPen(bak);
						
						last = next + len;
					}
					
					p->drawText(fxpos, baseline, h->m_text.mid(last, prevTextPos + rng.length() - last));
				}
				
				while ( cidx < cursor.count() )
				{
					const int xcoff = cursor.at(cidx) - prevTextPos;
					
					if ( xcoff < 0 )
					{
						++cidx;
						continue;
					}
					
					if ( xcoff < (rngIdx >= maxRange ? rngLen + 1 : rngLen) )
					{
						int xcpos = xpos;
						
						if ( xcoff )
						{
							if ( QDocumentPrivate::m_fixedPitch )
							{
								xcpos += xcoff * m_spaceWidth;
							} else {
								xcpos += p->fontMetrics().width(rng.left(xcoff));
							}
						}
						
						//qDebug("drawing cursor %i (col %i, x=%i)", cidx, cursor.at(cidx), xcpos);
						
						p->drawLine(xcpos, ypos, xcpos, ypos + QDocumentPrivate::m_lineSpacing);
						
						++cidx;
					} else {
						break;
					}
				}
				
				xpos += rwidth;
			}
			
			//qDebug("underline pos : %i", p->fontMetrics().underlinePos());
			
			if ( format.linescolor.isValid() )
				p->setPen(format.linescolor);
				
			const int ydo = qMin(baseline + p->fontMetrics().underlinePos(), ypos + QDocumentPrivate::m_lineSpacing - 1);
			const int yin = baseline - p->fontMetrics().strikeOutPos();
			const int yup = qMax(baseline - p->fontMetrics().overlinePos() + 1, ypos);
			
			if ( format.overline )
			{
				p->drawLine(xspos, yup, xpos, yup);
			}
			
			if ( format.strikeout )
			{
				p->drawLine(xspos, yin, xpos, yin);
			}
			
			if ( format.underline )
			{
				p->drawLine(xspos, ydo, xpos, ydo);
			}
			
			if ( format.waveUnderline )
 			{
				/*
				those goddamn font makers take liberties with common sense
				and make it so damn harder to figure proper y offset for wave
				underline (keeping the regular underline pos make it look
				weird or even invisible on some fonts... reference rendering
				with DejaVu Sans Mono)
				*/
				
				// we used fixed wave amplitude of 3 (well strictly speaking
				// amplitude would be 1.5 but let's forget about theory and
				// focus on getting that code to work...
				
				// gotta center things
				const int ycenter = ypos + QDocumentPrivate::m_lineSpacing - 3; 
									/*
									qMin(
										ypos + (QDocumentPrivate::m_ascent + QDocumentPrivate::m_lineSpacing) / 2,
										ypos + QDocumentPrivate::m_lineSpacing - 3
									);*/
				
				int cp = 0;
				brokenWave = false;
				
				while ( cp < rwidth )
 				{
					if ( !cp && !continuingWave )
					{
						dir = 0;
						p->drawLine(xspos, ycenter, xspos + 1, ycenter + 1);
						++cp;
					} else if ( !cp && brokenWave ) {
						if ( !dir )
							p->drawLine(xspos, ycenter, xspos + 1, ycenter + 1);
						else
							p->drawLine(xspos, ycenter, xspos + 1, ycenter - 1);
						
					} else {
						if ( cp + 2 > rwidth)
						{
							if ( !dir )
								p->drawLine(xspos + cp, ycenter - 1, xspos + cp + 1, ycenter);
							else
								p->drawLine(xspos + cp, ycenter + 1, xspos + cp + 1, ycenter);
							
							// trick to keep current direction
							dir ^= 1;
							
							brokenWave = true;
							++cp;
						} else {
							if ( !dir )
								p->drawLine(xspos + cp, ycenter - 1, xspos + cp + 2, ycenter + 1);
							else
								p->drawLine(xspos + cp, ycenter + 1, xspos + cp + 2, ycenter - 1);
							
							cp += 2;
						}
					}
					
					dir ^= 1;
 				}
				
				continuingWave = true;
 			} else {
				continuingWave = false;
				dir = 0;
			}
			
			p->setPen(oldpen);
		}
		
		if ( unbounded )
			p->fillRect(
							xpos, ypos,
							maxWidth - xpos, QDocumentPrivate::m_lineSpacing,
							pal.highlight()
						);
		
	}
}

void QDocumentPrivate::draw(QPainter *p, QDocument::PaintContext& cxt)
{
	QDocumentLineHandle *h;
	bool inSel = false, fullSel;
	QList<QDocumentCursorHandle*>::iterator cit;
	int i, realln, pos = 0, xOffset,
		firstLine = qMax(0, cxt.yoffset / m_lineSpacing),
		lastLine = qMax(0, firstLine + (cxt.height / m_lineSpacing));
		
	if ( cxt.height % m_lineSpacing )
		++lastLine;
		
	p->setFont(*m_font);
	
	QMarksManager *mm = QMarksManager::instance();
	
	const QMarkType& currentLineMark = mm->markType("current");
	
	QBrush bg, alternate,
		base = cxt.palette.base(),
		selbg = cxt.palette.highlight();
	
	if ( currentLineMark.gradient )
		alternate = QBrush(*currentLineMark.gradient);
	else
		alternate = cxt.palette.alternateBase();
	
	QVector<int> m_cursorLines(0), m_selectionBoundaries(0);
	
	int wrap = 0;
	i = textLine(firstLine, &wrap);
	firstLine -= wrap;
	realln = firstLine;
	
	//qDebug("lines [%i, %i]", firstLine, lastLine);
	
	pos += firstLine * m_lineSpacing;
	
	// adjust first line to take selections into account...
	foreach ( const QDocumentSelection& s, cxt.selections )
	{
		if ( (s.startLine < i) && (s.endLine >= i) )
		{
			inSel = true;
			break;
		}
	}
	
	const int numLines = lineCount();
	
	for ( ; realln <= lastLine; ++i )
	{
		if ( i >= numLines )
		{
			//qDebug("line %i not valid", i);
			break;
		}
		
		// TODO : use iterators with buffer
		h = at(i);
		
		if ( !h )
			break;
		
		const int wrap = m_constrained ? wrapBoundaries(h).count() : 0;
		const bool wrapped = wrap;
		
		//if ( wrapped )
		//	qDebug("line %i is wrapped over %i sublines", i, *wit);
		
		// selections stuff (must do it before whatever the visibility...)
		m_selectionBoundaries.clear();
		
		fullSel = false;
		
		if ( inSel )
			m_selectionBoundaries.prepend(0);
			
		foreach ( const QDocumentSelection& s, cxt.selections )
		{
			if ( i == s.startLine )
			{
				if ( !(m_selectionBoundaries.count() & 1) )
					m_selectionBoundaries.append(s.start);
					
				if ( i == s.endLine )
				{
					m_selectionBoundaries.append(s.end);
				} else {
					//++selLevel;
					inSel = true;
					//selEnd = h->m_text.length();
				}
			} else if ( inSel && (i == s.endLine) ) {
				
				if ( m_selectionBoundaries.count() % 2 )
					m_selectionBoundaries.append(s.end);
					
				//--selLevel;
				inSel = false;
			}
		}
		
		if ( inSel && m_selectionBoundaries.count() == 1 && m_selectionBoundaries.at(0) == 0 )
		{
			m_selectionBoundaries.clear();
			fullSel = true;
		}
		
		if ( h->hasFlag(QDocumentLine::Hidden) )
		{
			continue;
		} else
			++realln;
			
		if ( wrapped )
			realln += wrap;
			
		m_cursorLines.clear();
		
		bg = base;
		
		// idx = column = 0;
		xOffset = m_leftMargin; // margin
		
		// cursor(s) stuff
		cit = cxt.cursors.begin();
		
		int markPriority = 0;
		
		while ( cit != cxt.cursors.end() )
		{
			if ( (*cit)->lineNumber() == i )
			{
				if ( cxt.blinkingCursor )
					m_cursorLines.append((*cit)->columnNumber());
				
				if ( cxt.fillCursorRect )
				{
					markPriority = currentLineMark.priority;
					bg = alternate;
				}
				
				cit = cxt.cursors.erase(cit);
			} else {
				++cit;
			}
		}
		
		cit = cxt.extra.begin();
		
		while ( cit != cxt.extra.end() )
		{
			if ( (*cit)->lineNumber() == i )
			{
				m_cursorLines.append((*cit)->columnNumber());
				
				cit = cxt.extra.erase(cit);
			} else {
				++cit;
			}
		}
		
		qSort(m_cursorLines);
		
		QList<int> m = marks(h);
		
		// line marks stuff
		if ( m.count() )
		{
			const QMarkType& mt = mm->markType(mm->priority(m));
			
			if ( mt.gradient && mt.priority >= markPriority )
				bg = QBrush(*mt.gradient);
			
		}
		
		if ( realln < firstLine )
			continue;
			
		//qDebug("drawing line %i (visual %i)", i, realln);
		
		p->fillRect(qMax(cxt.xoffset, m_leftMargin), pos,
					cxt.width, m_lineSpacing * (wrap + 1), fullSel ? selbg : bg);
		
		p->save();
		
		// simplify line drawing
		p->translate(0, pos);
		
		// draw text
		draw(h, p, cxt.xoffset, cxt.width, m_selectionBoundaries, m_cursorLines, cxt.palette, fullSel);
		
		// see above
		p->translate(0, -pos);
		
		// draw fold rect indicator
		if ( h->hasFlag(QDocumentLine::CollapsedBlockStart) )
		{
			p->setBrush(Qt::NoBrush);
			p->setPen(QPen(Qt::blue, 1, Qt::DotLine));
			
			p->drawRect(m_leftMargin, pos,
						cxt.width - 4, m_lineSpacing * (wrap + 1) - 1);
						
		}
		
		p->restore();
		
		pos += m_lineSpacing;
		
		if ( wrapped )
		{
			pos += m_lineSpacing * wrap;
		}
		
		//qDebug("drawing line %i in %i ms", i, t.elapsed());
	}
	
	//qDebug("painting done"); // in %i ms...", t.elapsed());
}

QVector<int> QDocumentPrivate::wrapBoundaries(QDocumentLineHandle *h) const
{
	QHash<QDocumentLineHandle*, QVector<int> >::const_iterator it = m_boundaries.constFind(h);
	
	return it != m_boundaries.constEnd() ? *it : QVector<int>();
}

void QDocumentPrivate::setWidth(int width)
{
	bool oldConstraint = m_constrained;
	m_constrained = width > 0;
	
	if ( oldConstraint && m_constrained && m_width == width )
		return;
	
	if ( m_constrained )
	{
		//int oldWidth = m_width;
		
		m_width = width;
		
		// TODO : restore oldWidth < width optimization
		/*
		if ( oldConstraint && oldWidth < width )
		{
			// expand : simply remove old wraps if possible
			
			QMap<int, int>::iterator it = m_wrapped.begin();
			
			while ( it != m_wrapped.end() )
			{
				int ln = it.key();
				QDocumentLineHandle *h = at(ln);
				
				updateWrapBoundaries(h);
			}
		} else if ( oldWidth > width ) {
			// shrink : scan whole document and create new wraps wherever needed
			//qDebug("global width scan [constraint on]");
			//m_wrapped.clear();
			//m_boundaries.clear();
			setWidth();
		}
		*/
		setWidth();
	} else {
		//qDebug("global width scan [constraint off]");
		setWidth();
	}
	
	if ( m_editCursor )
	{
		m_editCursor->refreshColumnMemory();
	}
	
	//emitWidthChanged();
	setHeight();
	emit m_doc->lineCountChanged(lineCount());
	emit m_doc->widthConstraintChanged();
}

void QDocumentPrivate::setWidth()
{
	m_largest.clear();
	const int max = lineCount();
	
	if ( m_constrained )
	{
		//qDebug("constraining : %i", m_width);
		int i = 0;
		int first = -1;
		
		//m_boundaries.clear();
		
		QDocumentLineHandle *l;
		QDocumentBuffer::iterator dit = m_buffer->begin();
		
		while ( i < max )
		{
			l = dit.lineHandle();
			
			if ( !l )
				break;
			
			int span = 0;
			
			if ( updateWrapBoundaries(l, &span) )
			{
				// defer update of internal structure (third param)
				m_regions->setSpan(i, span, true);
			}
			
			dit.moveBy(1);
			++i;
		}
		
		m_regions->performDeferredUpdates();
		
		if ( first != -1 )
			emitFormatsChange(first, -1);
	} else {
		m_regions->clearSpans();
		
		int oldWidth = m_width;
		
		m_width = 0;
		
		QDocumentBuffer::iterator it = m_buffer->begin();
		
		while ( !it.atEnd() )
		{
			QDocumentLineHandle *l = it.lineHandle();
			it.moveBy(1);
			
			if ( l->hasFlag(QDocumentLine::Hidden) )
				continue;
				
			//l->m_boundaries.clear();
			
			if ( wrapBoundaries(l).count() )
			{
				if ( l->hasFlag(QDocumentLine::UseLayout) )
					l->setFlag(QDocumentLine::TextChanged, true);
				else
					l->setFlag(QDocumentLine::FormattingChanged, true);
			}
			
			int x, y;
			cursorToDocumentOffset(l, l->length(), x, y);
			
			if ( x > m_width )
			{
				m_width = x;
				
				m_largest.clear();
				m_largest << qMakePair(l, x);
			}
		}
		
		m_boundaries.clear();
		
		if ( m_width != oldWidth )
			emitWidthChanged();
	}
}

static const int widthCacheSize = 5;

void QDocumentPrivate::adjustWidth(int line)
{
	QDocumentLineHandle *l = at(line);
	
	if ( !l )
		return;
	
	if ( m_constrained )
	{
		int span = 0;
		
		if ( updateWrapBoundaries(l, &span) )
		{
			m_regions->setSpan(line, span);
			emitFormatsChange(line, -1);
			setHeight();
		}
	} else {
		
		int x, y;
		cursorToDocumentOffset(l, l->length(), x, y);
		
		if ( x > m_width )
		{
			m_width = x;
			emitWidthChanged();
			
			m_largest.clear();
			m_largest << qMakePair(l, x);
		} else if ( m_largest.count() && (m_largest.at(0).first == l) ) {
			int old = m_largest.at(0).second;
			
			if ( x < old )
				setWidth();
		}
	}
}

void QDocumentPrivate::setHeight()
{
	int oldHeight = m_height;
	const int numLines = lineCount();
	int last = m_regions->visualLineCount();//visualLine(numLines - 1) + 1;
	
	if ( numLines )
		last += wrapBoundaries(m_buffer->last()).count();
	
	//qDebug("setHeight : %i lines (%i visual)", numLines, last);
	
	m_height = last * m_lineSpacing;
	
	if ( oldHeight != m_height )
		emitHeightChanged();
}

void QDocumentPrivate::setFont(const QFont& f)
{
	if ( !m_font )
	{
		m_font = new QFont;
		m_fontMetrics = new QFontMetrics(*m_font);
	}
	
	*m_font = f;
	
	// set the style hint so that if the font is not found a Courier (fixed pitch) one will be used
	m_font->setStyleHint(QFont::Courier, QFont::PreferQuality);
	
	*m_fontMetrics = QFontMetrics(*m_font);
	
	m_spaceWidth = m_fontMetrics->width(' ');
	m_lineSpacing = m_fontMetrics->lineSpacing();
	m_ascent = m_fontMetrics->ascent();
	m_descent = m_fontMetrics->descent();
	m_leading = m_fontMetrics->leading();
	
	m_lineHeight = m_fontMetrics->height();
	//m_lineHeight = m_ascent + m_descent - 2;
	
	m_fixedPitch = QFontInfo(*m_font).fixedPitch();
	
	//if ( !m_fixedPitch )
	//	qDebug("unsafe computations...");
	
	foreach ( QDocumentPrivate *d, m_documents )
	{
		d->updateFormatCache();
		d->setWidth();
		d->setHeight();
	}
}

void QDocumentPrivate::setFormatScheme(QFormatScheme *f)
{
	m_formatScheme = f;
	updateFormatCache();
}

void QDocumentPrivate::tunePainter(QPainter *p, int fid)
{
	if ( fid < m_fonts.count() )
	{
		p->setFont(m_fonts.at(fid));
		//p->setPen(m_colors.at(fid));
	} else {
		QFormat f = m_formatScheme->format(fid);
		QFont font(*m_font);
		font.setWeight(f.weight);
		font.setItalic(f.italic);
		p->setFont(font);
		//p->setPen(Qt::black);
	}
}

void QDocumentPrivate::updateFormatCache()
{
	m_fonts.clear();
	
	if ( !m_font )
		return;
	
	if ( !m_formatScheme )
	{
		m_fonts << *m_font;
		return;
	}
	
	QFont f(*m_font);
	/*
	const int end = m_formatScheme->formatCount();
	
	m_fonts.reserve(end);
	
	for ( int i = 0; i < end; i++ )
	{
		QFormat fmt = m_formatScheme->format(i);
		
		f.setWeight(fmt.weight);
		f.setItalic(fmt.italic);
		
		m_fonts << f;
	}
	*/
	//foreach ( QDocumentPrivate *d, m_documents )
	//	d->emitFormatsChanged();
	
	emitFormatsChanged();
}

void QDocumentPrivate::emitWidthChanged()
{
	if ( !m_doc )
		return;
		
	emit m_doc->widthChanged(m_width);
	
	emit m_doc->sizeChanged(QSize(m_width, m_height));
}

void QDocumentPrivate::emitHeightChanged()
{
	if ( !m_doc )
		return;
		
	emit m_doc->heightChanged(m_height);
	
	emit m_doc->sizeChanged(QSize(m_width, m_height));
}

void QDocumentPrivate::insertLines(int after, const QVector<QDocumentLineHandle*>& l)
{
	//qDebug("inserting : %i, %i", after, l.count());
	
	++after;
	
	m_buffer->insertLines(after, l);
	
	int line = after;
	const quint8 flagMask = ~(QDocumentLine::CollapsedBlockStart | QDocumentLine::CollapsedBlockEnd);
	
	foreach ( QDocumentLineHandle *h, l )
	{
		h->m_state &= flagMask;
		h->m_state |= QDocumentLine::TextChanged | QDocumentLine::FormattingChanged;
		
		if ( m_constrained )
		{
			updateWrapBoundaries(h);
		} else {
			
			int x, y;
			cursorToDocumentOffset(h, h->length(), x, y);
			
			if ( x > m_width )
			{
				m_width = x;
				emitWidthChanged();
				
				m_largest.clear();
				m_largest << qMakePair(h, x);
			} else if ( m_largest.count() && (m_largest.at(0).first == h) ) {
				int old = m_largest.at(0).second;
				
				if ( x < old )
					setWidth();
			}
		}
		
		++line;
	}
	
	m_regions->linesInserted(after, l.count());
	emit m_doc->lineCountChanged(lineCount());
	
	setHeight();
}

void QDocumentPrivate::removeLines(int after, int n)
{
	// TODO : evaluate usefulness
	#if 0
	if ( (after >= 0) && (after < m_lines.count()) )
		m_lines.at(after)->setFlag(QDocumentLine::CollapsedBlockStart, false);
	#endif
	
	++after;
	
	int count = n;
	QDocumentBuffer::iterator it = m_buffer->find(after);
	
	do
	{
		m_boundaries.remove(it.lineHandle());
		it.moveBy(1);
	} while ( --count > 0 );
	
	m_buffer->removeLines(after, n);
	
	m_regions->linesRemoved(after, n);
	
	emit m_doc->lineCountChanged(lineCount());
	setHeight();
}

QDocumentLineHandle* QDocumentPrivate::at(int line) const
{
	return m_buffer->at(line);
}

QDocumentBuffer::iterator QDocumentPrivate::begin() const
{
	return m_buffer->begin();
}

QDocumentBuffer::iterator QDocumentPrivate::end() const
{
	return m_buffer->end();
}

QDocumentBuffer::iterator QDocumentPrivate::find(int line) const
{
	return m_buffer->find(line);
}

int QDocumentPrivate::indexOf(const QDocumentLineHandle *l) const
{
	return m_buffer->indexOf(l);
}

void QDocumentPrivate::beginChangeBlock()
{
	//qDebug("<macro>");
	m_commands.beginMacro(QString());
}

void QDocumentPrivate::endChangeBlock()
{
	m_commands.endMacro();
	//qDebug("</macro>");
}

/*!
	\brief Acquire group id
*/
int QDocumentPrivate::getNextGroupId()
{
	if ( m_freeGroupIds.count() )
		return m_freeGroupIds.takeFirst();
	
	return ++m_lastGroupId;
}

/*!
	\brief Relase group id
*/
void QDocumentPrivate::releaseGroupId(int groupId)
{
	if ( groupId == m_lastGroupId )
	{
		--m_lastGroupId;
		while ( m_freeGroupIds.removeAll(m_lastGroupId) )
		{
			--m_lastGroupId;
		}
	} else {
		m_freeGroupIds << groupId;
	}
}

/*!
	\brief Clear matches
*/
void QDocumentPrivate::clearMatches(int groupId)
{
	QHash<int, MatchList>::iterator mit = m_matches.find(groupId);
	
	if ( mit == m_matches.end() )
	{
		return;
	}
	
	MatchList& matches = *mit;
	
	foreach ( const Match& m, matches )
	{
		m.h->removeOverlay(m.range);
	}
	
	matches.index = matches.count();
}

/*!
	\brief Highlight the matched sequences
	
	\note Both position are BEFORE the matched characters (cursor position).
*/
void QDocumentPrivate::addMatch(int groupId, int line, int pos, int len, int format)
{
	//qDebug("match (%i, %i, %i)", line, pos, len);
	
	Match m;
	m.line = line;
	m.h = at(line);
	m.range = QFormatRange(pos, len, format);
	m_matches[groupId] << m;
	
	m.h->addOverlay(m.range);
}

void QDocumentPrivate::flushMatches(int groupId)
{
	QHash<int, MatchList>::iterator mit = m_matches.find(groupId);
	
	if ( mit == m_matches.end() )
	{
		return;
	}
	
	MatchList& matches = *mit;
	
	QMap<int, int> areas;
	
	foreach ( const Match& m, matches )
	{
		int n = 1;
		int l = m.line;
		
		//qDebug("simple:(%i, %i)", l, 1);
		
		QMap<int, int>::iterator tmp, it = areas.find(l);
		
		if ( it != areas.end() )
			continue;
			
		it = areas.insert(m.line, n);
		
		if ( it != areas.end() && (it - 1) != areas.end() )
		{
			tmp = it - 1;
			int off = tmp.key() + *tmp - l;
			
			if ( off >= 0 && (off < n) )
			{
				*tmp += n - off;
				it = areas.erase(it) - 1;
			}
		}
		
		if ( it != areas.end() && (it + 1) != areas.end() )
		{
			tmp = it + 1;
			int off = it.key() + *it - tmp.key();
			
			if ( off >= 0 && (off < *tmp) )
			{
				*it += *tmp;
				areas.erase(tmp);
			}
		}
		//emitFormatsChange(m.line, 1);
	}
	
	// remove old matches
	while ( matches.index )
	{
		matches.removeFirst();
		--matches.index;
	}
	
	// send update messages
	QMap<int, int>::const_iterator it = areas.constBegin();
	
	while ( it != areas.constEnd() )
	{
		//qDebug("merged:(%i, %i)", it.key(), *it);
		emitFormatsChange(it.key(), *it);
		
		++it;
	}
	
	// update storage "meta-data"
	if ( matches.isEmpty() )
	{
		m_matches.remove(groupId);
		
		releaseGroupId(groupId);
	}
	//qDebug("done with matches");
}

QList<int> QDocumentPrivate::marks(QDocumentLineHandle *h) const
{
	QList<int> l;
	QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator it = m_marksLookup.constFind(h);
	
	if ( it != m_marksLookup.constEnd() )
	{
		const QList<QDocumentMark*>& ml = *it;
		
		foreach ( QDocumentMark *m, ml )
			l << m->type();
	}
	
	return l;
}

void QDocumentPrivate::addMark(QDocumentMark *m)
{
	m_marks << m;
	
	QList<QDocumentLineHandle*> lines = m->location()->selectionSpan();
	
	foreach ( QDocumentLineHandle *l, lines )
	{
		QList<QDocumentMark*>& lm = m_marksLookup[l];
		
		lm << m;
		m_maxMarksPerLine = qMax(lm.count(), m_maxMarksPerLine);
	}
	
	emitMarkAdded(m);
}

void QDocumentPrivate::removeMark(QDocumentMark *m)
{
	int n = m_marks.removeAll(m);
	
	if ( !n )
		return;
	
	// TODO : be smarter and use location instead...
	QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::iterator it = m_marksLookup.begin();
	const QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::iterator end = m_marksLookup.end();
	
	bool recompMax = false;
	
	while ( it != end )
	{
		QList<QDocumentMark*>& l = *it;
		int count = l.count();
		
		n = l.removeAll(m);
		
		if ( n && !recompMax )
			recompMax = (count == m_maxMarksPerLine);
		
		++it;
	}
	
	if ( recompMax )
	{
		QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator
			rit = m_marksLookup.constBegin(),
			end = m_marksLookup.constEnd();
			
		m_maxMarksPerLine = 0;
		
		while ( rit != end )
		{
			m_maxMarksPerLine = qMax(rit->count(), m_maxMarksPerLine);
			++rit;
		}
	}
	
	emitMarkRemoved(m);
}

void QDocumentPrivate::addMark(int line, int mid)
{
	QDocumentLineHandle *lh = at(line);
	QDocumentCursorHandle *ch = new QDocumentCursorHandle(m_doc, line);
	
	QList<QDocumentMark*>& l = m_marksLookup[lh];
	
	QDocumentMark *m = new QDocumentMark(ch, mid);
	
	l << m;
	m_marks << m;
	
	m_maxMarksPerLine = qMax(l.count(), m_maxMarksPerLine);
	
	//emitMarkChanged(lh, mid, true);
	emitMarkAdded(m);
}

void QDocumentPrivate::toggleMark(int line, int mid)
{
	QDocumentLineHandle *lh = at(line);
	
	if ( marks(lh).contains(mid) )
	{
		removeMark(line, mid);
	} else {
		addMark(line, mid);
	}
}

void QDocumentPrivate::removeMark(int line, int mid)
{
	//QDocumentLineHandle *h = at(line);
	
	removeMark(at(line), mid);
}

void QDocumentPrivate::addMark(QDocumentLineHandle *h, int mid)
{
	addMark(indexOf(h), mid);
}

void QDocumentPrivate::toggleMark(QDocumentLineHandle *h, int mid)
{
	if ( marks(h).contains(mid) )
	{
		removeMark(h, mid);
	} else {
		addMark(h, mid);
	}
}

void QDocumentPrivate::removeMark(QDocumentLineHandle *h, int mid)
{
	QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::iterator it = m_marksLookup.find(h);
	
	if ( it == m_marksLookup.end() )
		return;
	
	QList<QDocumentMark*>& ml = *it;
	int n = 0, i = 0, count = ml.count();
	
	while ( i < count )
	{
		QDocumentMark *m = ml.at(i);
		
		if ( m->type() != mid )
		{
			++i;
			continue;
		}
		
		if ( m->cursor().hasSelection() )
		{
			// discard : such marks must have been added with "advanced" API and must be managed with it
			++i;
		} else {
			ml.removeAt(i);
			m_marks.removeAll(m);
			
			emitMarkRemoved(m);
			
			// removing through simple API causes (defered?) mark deletion?
			//delete m;
			
			--count;
			++n;
		}
	}
	
	if ( ml.isEmpty() )
		m_marksLookup.erase(it);
	
	if ( n && (count == m_maxMarksPerLine) )
	{
		QHash<QDocumentLineHandle*, QList<QDocumentMark*> >::const_iterator
			rit = m_marksLookup.constBegin(),
			end = m_marksLookup.constEnd();
			
		m_maxMarksPerLine = 0;
		
		while ( rit != end )
		{
			m_maxMarksPerLine = qMax(rit->count(), m_maxMarksPerLine);
			++rit;
		}
	}
	
	//emitMarkChanged(h, mid, false);
}

QList<QDocumentMark*> QDocumentPrivate::findMarks(QDocumentCursorHandle *h, const QList<int>& ids) const
{
	QList<QDocumentMark*> result;
	
	if ( h )
	{
		// search on line / within selection
		
		QList<QDocumentLineHandle*> sel = h->selectionSpan();
		
		foreach ( QDocumentLineHandle *l, sel )
		{
			QList<QDocumentMark*> lm = m_marksLookup.value(l);
			
			foreach ( QDocumentMark *m, lm )
			{
				if ( m->flags() & QDocumentMark::ColumnMatters )
				{
					if ( h->intersect(m->location()) )
						continue;
				}
				
				if ( ids.isEmpty() || ids.contains(m->type()) )
					result << m;
			}
		}
	} else {
		// search whole document
		
		foreach ( QDocumentMark *m, m_marks )
		{
			if ( ids.isEmpty() || ids.contains(m->type()) )
				result << m;
		}
	}
	
	return result;
}

int QDocumentPrivate::visualLine(int textLine) const
{
	if ( textLine < 0 )
		return 0;
	
	return m_regions->visualLineNumber(textLine);
}

int QDocumentPrivate::textLine(int visualLine, int *wrap) const
{
	if ( visualLine < 0 )
		return 0;
	
	
	return m_regions->logicalLineNumber(visualLine, wrap);
}

void QDocumentPrivate::hideEvent(int line, int count)
{
	m_regions->hide(line, count);
	
	setHeight();
	//emitFormatsChange(line, count);
	emitFormatsChanged();
}

void QDocumentPrivate::showEvent(int line, int count)
{
	m_regions->show(line, count);
	
	setHeight();
	//emitFormatsChange(line, count);
	emitFormatsChanged();
}

void QDocumentPrivate::emitFormatsChange(int line, int lines)
{
	emit m_doc->formatsChange(line, lines);
}

void QDocumentPrivate::emitContentsChange(int line, int lines)
{
	int n = lines;
	
	if ( m_language )
	{
		n = m_language->tokenize(m_doc, line, lines);
	}
	//qDebug("%i, %i : %i", line, lines, n);
	
	emit m_doc->contentsChange(line, lines);
	emit m_doc->contentsChanged();
	
	if ( n > lines )
		emitFormatsChange(line + lines, n - lines);
}

void QDocumentPrivate::emitFormatsChanged()
{
	emit m_doc->formatsChanged();
}

void QDocumentPrivate::emitContentsChanged()
{
	//emit m_doc->contentsChanged();
}

void QDocumentPrivate::emitLineDeleted(QDocumentLineHandle *h)
{
	if ( !m_deleting )
	{
		m_status.remove(h);
		m_boundaries.remove(h);
		m_marksLookup.remove(h);
	}
	
	emit m_doc->lineDeleted(h);
}

void QDocumentPrivate::emitMarkAdded(QDocumentMark *m)
{
	emitFormatsChanged();
	emit m_doc->markChanged(m);
}

void QDocumentPrivate::emitMarkRemoved(QDocumentMark *m)
{
	emitFormatsChanged();
	emit m_doc->markChanged(m);
}

void QDocumentPrivate::emitMarkChanged(QDocumentMark *m)
{
	emitFormatsChanged();
	emit m_doc->markChanged(m);
}

/*! @} */
